from .alg import *
from .colors import *
from .conf import *
from .doc import *
from .dtime import *
from .fs import *
from .log import *
from .misc import *
from .net import *
from .paral import *
from .pkg import *
from .proc import *
from .ser import *
from .str import *
from .sys_ import *
from .test import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/19 17:21:57
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from collections import OrderedDict
from ..str import dict_to_str
from ..fs import file_read_lines, file_write_lines
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['NameIndexFile', 'cls_to_name']

# 根据配置文件获取索引
class NameIndexFile:
    def __init__(self, cls_file = None):
        self.cls_data_ = OrderedDict()
        if cls_file:
            self.setClsFile(cls_file)

    def setClsFile(self, cls_file):
        for i,line in enumerate(file_read_lines(cls_file)):
            self.cls_data_[i] = line
        return self

    def getName(self, idx):
        _idx = int(idx)
        return self.cls_data_.get(_idx, f'Unknow[{idx}]')

    def getIndex(self, name):
        if name:
            for _idx, _name in self.cls_data_.items():
                if name in _name:
                    return _idx
        return -1

    def getData(self):
        return self.cls_data_

    def __str__(self):
        return dict_to_str(self.cls_data_, start=20, stop=20)

def cls_to_name(index, filename):
    nameIdx = NameIndexFile(filename)
    return nameIdx.getName(index)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/20 18:31:49

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
import os, shutil
import laok
from ..fs import path_parent
from ..proc import workdir_scope
# ===============================================================================
r'''
'''
# ===============================================================================

def pypi_upload(lib_dir=None, rm_temp=True, upload = True):
    if not lib_dir:
        lib_dir = path_parent(__file__, 4)

    print(f'lib_dir = {lib_dir}')

    # 切换到 lib_dir
    with workdir_scope(lib_dir):

        # 编译工具包
        os.system('python setup.py sdist')

        # 上传到pypi
        if upload:
            os.system('twine upload dist/*.gz')

        if rm_temp:  #删除临时文件
            for rmdir in ['build', 'dist', 'laok.egg-info']:
                shutil.rmtree(rmdir, ignore_errors=True)


""" Helper methods for loading and parsing KITTI data.

Author: Charles R. Qi, Kui Xu
Date: September 2017/2018
"""

import numpy as np
import cv2
import os, math
from scipy.optimize import leastsq
from PIL import Image

TOP_Y_MIN = -30
TOP_Y_MAX = +30
TOP_X_MIN = 0
TOP_X_MAX = 100
TOP_Z_MIN = -3.5
TOP_Z_MAX = 0.6

TOP_X_DIVISION = 0.2
TOP_Y_DIVISION = 0.2
TOP_Z_DIVISION = 0.3

cbox = np.array([[0, 70.4], [-40, 40], [-3, 2]])


class Object2d(object):
    """ 2d object label """

    def __init__(self, label_file_line):
        data = label_file_line.split(" ")

        # extract label, truncation, occlusion
        self.img_name = int(data[0])  # 'Car', 'Pedestrian', ...
        self.typeid = int(data[1])  # truncated pixel ratio [0..1]
        self.prob = float(data[2])
        self.box2d = np.array([int(data[3]), int(data[4]), int(data[5]), int(data[6])])

    def print_object(self):
        print(
            "img_name, typeid, prob: %s, %d, %f"
            % (self.img_name, self.typeid, self.prob)
        )
        print(
            "2d bbox (x0,y0,x1,y1): %d, %d, %d, %d"
            % (self.box2d[0], self.box2d[1], self.box2d[2], self.box2d[3])
        )


class Object3d(object):
    """ 3d object label """

    def __init__(self, label_file_line):
        data = label_file_line.split(" ")
        data[1:] = [float(x) for x in data[1:]]

        # extract label, truncation, occlusion
        self.type = data[0]  # 'Car', 'Pedestrian', ...
        self.truncation = data[1]  # truncated pixel ratio [0..1]
        self.occlusion = int(
            data[2]
        )  # 0=visible, 1=partly occluded, 2=fully occluded, 3=unknown
        self.alpha = data[3]  # object observation angle [-pi..pi]

        # extract 2d bounding box in 0-based coordinates
        self.xmin = data[4]  # left
        self.ymin = data[5]  # top
        self.xmax = data[6]  # right
        self.ymax = data[7]  # bottom
        self.box2d = np.array([self.xmin, self.ymin, self.xmax, self.ymax])

        # extract 3d bounding box information
        self.h = data[8]  # box height
        self.w = data[9]  # box width
        self.l = data[10]  # box length (in meters)
        self.t = (data[11], data[12], data[13])  # location (x,y,z) in camera coord.
        self.ry = data[14]  # yaw angle (around Y-axis in camera coordinates) [-pi..pi]

    def estimate_diffculty(self):
        """ Function that estimate difficulty to detect the object as defined in kitti website"""
        # height of the bounding box
        bb_height = np.abs(self.xmax - self.xmin)

        if bb_height >= 40 and self.occlusion == 0 and self.truncation <= 0.15:
            return "Easy"
        elif bb_height >= 25 and self.occlusion in [0, 1] and self.truncation <= 0.30:
            return "Moderate"
        elif (
            bb_height >= 25 and self.occlusion in [0, 1, 2] and self.truncation <= 0.50
        ):
            return "Hard"
        else:
            return "Unknown"

    def print_object(self):
        print(
            "Type, truncation, occlusion, alpha: %s, %d, %d, %f"
            % (self.type, self.truncation, self.occlusion, self.alpha)
        )
        print(
            "2d bbox (x0,y0,x1,y1): %f, %f, %f, %f"
            % (self.xmin, self.ymin, self.xmax, self.ymax)
        )
        print("3d bbox h,w,l: %f, %f, %f" % (self.h, self.w, self.l))
        print(
            "3d bbox location, ry: (%f, %f, %f), %f"
            % (self.t[0], self.t[1], self.t[2], self.ry)
        )
        print("Difficulty of estimation: {}".format(self.estimate_diffculty()))


class Calibration(object):
    """ Calibration matrices and utils
        3d XYZ in <label>.txt are in rect camera coord.
        2d box xy are in image2 coord
        Points in <lidar>.bin are in Velodyne coord.

        y_image2 = P^2_rect * x_rect
        y_image2 = P^2_rect * R0_rect * Tr_velo_to_cam * x_velo
        x_ref = Tr_velo_to_cam * x_velo
        x_rect = R0_rect * x_ref

        P^2_rect = [f^2_u,  0,      c^2_u,  -f^2_u b^2_x;
                    0,      f^2_v,  c^2_v,  -f^2_v b^2_y;
                    0,      0,      1,      0]
                 = K * [1|t]

        image2 coord:
         ----> x-axis (u)
        |
        |
        v y-axis (v)

        velodyne coord:
        front x, left y, up z

        rect/ref camera coord:
        right x, down y, front z

        Ref (KITTI paper): http://www.cvlibs.net/publications/Geiger2013IJRR.pdf

        TODO(rqi): do matrix multiplication only once for each projection.
    """

    def __init__(self, calib_filepath, from_video=False):
        if from_video:
            calibs = self.read_calib_from_video(calib_filepath)
        else:
            calibs = self.read_calib_file(calib_filepath)
        # Projection matrix from rect camera coord to image2 coord
        self.P = calibs["P2"]
        self.P = np.reshape(self.P, [3, 4])
        # Rigid transforms from Velodyne coord to reference camera coord
        self.V2C = calibs["Tr_velo_to_cam"]
        self.V2C = np.reshape(self.V2C, [3, 4])
        self.C2V = inverse_rigid_trans(self.V2C)
        # Rotation from reference camera coord to rect camera coord
        self.R0 = calibs["R0_rect"]
        self.R0 = np.reshape(self.R0, [3, 3])

        # Camera intrinsics and extrinsics
        self.c_u = self.P[0, 2]
        self.c_v = self.P[1, 2]
        self.f_u = self.P[0, 0]
        self.f_v = self.P[1, 1]
        self.b_x = self.P[0, 3] / (-self.f_u)  # relative
        self.b_y = self.P[1, 3] / (-self.f_v)

    def read_calib_file(self, filepath):
        """ Read in a calibration file and parse into a dictionary.
        Ref: https://github.com/utiasSTARS/pykitti/blob/master/pykitti/utils.py
        """
        data = {}
        with open(filepath, "r") as f:
            for line in f.readlines():
                line = line.rstrip()
                if len(line) == 0:
                    continue
                key, value = line.split(":", 1)
                # The only non-float values in these files are dates, which
                # we don't care about anyway
                try:
                    data[key] = np.array([float(x) for x in value.split()])
                except ValueError:
                    pass

        return data

    def read_calib_from_video(self, calib_root_dir):
        """ Read calibration for camera 2 from video calib files.
            there are calib_cam_to_cam and calib_velo_to_cam under the calib_root_dir
        """
        data = {}
        cam2cam = self.read_calib_file(
            os.path.join(calib_root_dir, "calib_cam_to_cam.txt")
        )
        velo2cam = self.read_calib_file(
            os.path.join(calib_root_dir, "calib_velo_to_cam.txt")
        )
        Tr_velo_to_cam = np.zeros((3, 4))
        Tr_velo_to_cam[0:3, 0:3] = np.reshape(velo2cam["R"], [3, 3])
        Tr_velo_to_cam[:, 3] = velo2cam["T"]
        data["Tr_velo_to_cam"] = np.reshape(Tr_velo_to_cam, [12])
        data["R0_rect"] = cam2cam["R_rect_00"]
        data["P2"] = cam2cam["P_rect_02"]
        return data

    def cart2hom(self, pts_3d):
        """ Input: nx3 points in Cartesian
            Oupput: nx4 points in Homogeneous by pending 1
        """
        n = pts_3d.shape[0]
        pts_3d_hom = np.hstack((pts_3d, np.ones((n, 1))))
        return pts_3d_hom

    # ===========================
    # ------- 3d to 3d ----------
    # ===========================
    def project_velo_to_ref(self, pts_3d_velo):
        pts_3d_velo = self.cart2hom(pts_3d_velo)  # nx4
        return np.dot(pts_3d_velo, np.transpose(self.V2C))

    def project_ref_to_velo(self, pts_3d_ref):
        pts_3d_ref = self.cart2hom(pts_3d_ref)  # nx4
        return np.dot(pts_3d_ref, np.transpose(self.C2V))

    def project_rect_to_ref(self, pts_3d_rect):
        """ Input and Output are nx3 points """
        return np.transpose(np.dot(np.linalg.inv(self.R0), np.transpose(pts_3d_rect)))

    def project_ref_to_rect(self, pts_3d_ref):
        """ Input and Output are nx3 points """
        return np.transpose(np.dot(self.R0, np.transpose(pts_3d_ref)))

    def project_rect_to_velo(self, pts_3d_rect):
        """ Input: nx3 points in rect camera coord.
            Output: nx3 points in velodyne coord.
        """
        pts_3d_ref = self.project_rect_to_ref(pts_3d_rect)
        return self.project_ref_to_velo(pts_3d_ref)

    def project_velo_to_rect(self, pts_3d_velo):
        pts_3d_ref = self.project_velo_to_ref(pts_3d_velo)
        return self.project_ref_to_rect(pts_3d_ref)

    # ===========================
    # ------- 3d to 2d ----------
    # ===========================
    def project_rect_to_image(self, pts_3d_rect):
        """ Input: nx3 points in rect camera coord.
            Output: nx2 points in image2 coord.
        """
        pts_3d_rect = self.cart2hom(pts_3d_rect)
        pts_2d = np.dot(pts_3d_rect, np.transpose(self.P))  # nx3
        pts_2d[:, 0] /= pts_2d[:, 2]
        pts_2d[:, 1] /= pts_2d[:, 2]
        return pts_2d[:, 0:2]

    def project_velo_to_image(self, pts_3d_velo):
        """ Input: nx3 points in velodyne coord.
            Output: nx2 points in image2 coord.
        """
        pts_3d_rect = self.project_velo_to_rect(pts_3d_velo)
        return self.project_rect_to_image(pts_3d_rect)

    def project_8p_to_4p(self, pts_2d):
        x0 = np.min(pts_2d[:, 0])
        x1 = np.max(pts_2d[:, 0])
        y0 = np.min(pts_2d[:, 1])
        y1 = np.max(pts_2d[:, 1])
        x0 = max(0, x0)
        # x1 = min(x1, proj.image_width)
        y0 = max(0, y0)
        # y1 = min(y1, proj.image_height)
        return np.array([x0, y0, x1, y1])

    def project_velo_to_4p(self, pts_3d_velo):
        """ Input: nx3 points in velodyne coord.
            Output: 4 points in image2 coord.
        """
        pts_2d_velo = self.project_velo_to_image(pts_3d_velo)
        return self.project_8p_to_4p(pts_2d_velo)

    # ===========================
    # ------- 2d to 3d ----------
    # ===========================
    def project_image_to_rect(self, uv_depth):
        """ Input: nx3 first two channels are uv, 3rd channel
                   is depth in rect camera coord.
            Output: nx3 points in rect camera coord.
        """
        n = uv_depth.shape[0]
        x = ((uv_depth[:, 0] - self.c_u) * uv_depth[:, 2]) / self.f_u + self.b_x
        y = ((uv_depth[:, 1] - self.c_v) * uv_depth[:, 2]) / self.f_v + self.b_y
        pts_3d_rect = np.zeros((n, 3))
        pts_3d_rect[:, 0] = x
        pts_3d_rect[:, 1] = y
        pts_3d_rect[:, 2] = uv_depth[:, 2]
        return pts_3d_rect

    def project_image_to_velo(self, uv_depth):
        pts_3d_rect = self.project_image_to_rect(uv_depth)
        return self.project_rect_to_velo(pts_3d_rect)

    def project_depth_to_velo(self, depth, constraint_box=True):
        depth_pt3d = get_depth_pt3d(depth)
        depth_UVDepth = np.zeros_like(depth_pt3d)
        depth_UVDepth[:, 0] = depth_pt3d[:, 1]
        depth_UVDepth[:, 1] = depth_pt3d[:, 0]
        depth_UVDepth[:, 2] = depth_pt3d[:, 2]
        # print("depth_pt3d:",depth_UVDepth.shape)
        depth_pc_velo = self.project_image_to_velo(depth_UVDepth)
        # print("dep_pc_velo:",depth_pc_velo.shape)
        if constraint_box:
            depth_box_fov_inds = (
                (depth_pc_velo[:, 0] < cbox[0][1])
                & (depth_pc_velo[:, 0] >= cbox[0][0])
                & (depth_pc_velo[:, 1] < cbox[1][1])
                & (depth_pc_velo[:, 1] >= cbox[1][0])
                & (depth_pc_velo[:, 2] < cbox[2][1])
                & (depth_pc_velo[:, 2] >= cbox[2][0])
            )
            depth_pc_velo = depth_pc_velo[depth_box_fov_inds]
        return depth_pc_velo


def get_depth_pt3d(depth):
    pt3d = []
    for i in range(depth.shape[0]):
        for j in range(depth.shape[1]):
            pt3d.append([i, j, depth[i, j]])
    return np.array(pt3d)


def rotx(t):
    """ 3D Rotation about the x-axis. """
    c = np.cos(t)
    s = np.sin(t)
    return np.array([[1, 0, 0], [0, c, -s], [0, s, c]])


def roty(t):
    """ Rotation about the y-axis. """
    c = np.cos(t)
    s = np.sin(t)
    return np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]])


def rotz(t):
    """ Rotation about the z-axis. """
    c = np.cos(t)
    s = np.sin(t)
    return np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])


def transform_from_rot_trans(R, t):
    """ Transforation matrix from rotation matrix and translation vector. """
    R = R.reshape(3, 3)
    t = t.reshape(3, 1)
    return np.vstack((np.hstack([R, t]), [0, 0, 0, 1]))


def inverse_rigid_trans(Tr):
    """ Inverse a rigid body transforms matrix (3x4 as [R|t])
        [R'|-R't; 0|1]
    """
    inv_Tr = np.zeros_like(Tr)  # 3x4
    inv_Tr[0:3, 0:3] = np.transpose(Tr[0:3, 0:3])
    inv_Tr[0:3, 3] = np.dot(-np.transpose(Tr[0:3, 0:3]), Tr[0:3, 3])
    return inv_Tr


def read_label(label_filename):
    lines = [line.rstrip() for line in open(label_filename)]
    objects = [Object3d(line) for line in lines]
    return objects


def load_image(img_filename):
    return cv2.imread(img_filename)


def load_depth_v(img_filename):
    # return cv2.imread(img_filename)
    disp_img = cv2.imread(img_filename, cv2.IMREAD_UNCHANGED)
    disp_img = disp_img.astype(np.float)
    return disp_img / 256.0


def load_depth0(img_filename):
    # return cv2.imread(img_filename)
    depth_img = np.array(Image.open(img_filename), dtype=int)

    depth_img = depth_img.astype(np.float) / 256.0

    return depth_img


def load_depth(img_filename):
    isexist = True
    disp_img = cv2.imread(img_filename, cv2.IMREAD_UNCHANGED)
    if disp_img is None:
        isexist = False
        disp_img = np.zeros((370, 1224))
    else:
        disp_img = disp_img.astype(np.float)
    return disp_img / 256.0, isexist


def load_velo_scan(velo_filename, dtype=np.float32, n_vec=4):
    scan = np.fromfile(velo_filename, dtype=dtype)
    scan = scan.reshape((-1, n_vec))
    return scan


def lidar_to_top_coords(x, y):
    # print("TOP_X_MAX-TOP_X_MIN:",TOP_X_MAX,TOP_X_MIN)
    Xn = int((TOP_X_MAX - TOP_X_MIN) // TOP_X_DIVISION) + 1
    Yn = int((TOP_Y_MAX - TOP_Y_MIN) // TOP_Y_DIVISION) + 1
    xx = Yn - int((y - TOP_Y_MIN) // TOP_Y_DIVISION)
    yy = Xn - int((x - TOP_X_MIN) // TOP_X_DIVISION)

    return xx, yy


def lidar_to_top(lidar):

    idx = np.where(lidar[:, 0] > TOP_X_MIN)
    lidar = lidar[idx]
    idx = np.where(lidar[:, 0] < TOP_X_MAX)
    lidar = lidar[idx]

    idx = np.where(lidar[:, 1] > TOP_Y_MIN)
    lidar = lidar[idx]
    idx = np.where(lidar[:, 1] < TOP_Y_MAX)
    lidar = lidar[idx]

    idx = np.where(lidar[:, 2] > TOP_Z_MIN)
    lidar = lidar[idx]
    idx = np.where(lidar[:, 2] < TOP_Z_MAX)
    lidar = lidar[idx]

    pxs = lidar[:, 0]
    pys = lidar[:, 1]
    pzs = lidar[:, 2]
    prs = lidar[:, 3]
    qxs = ((pxs - TOP_X_MIN) // TOP_X_DIVISION).astype(np.int32)
    qys = ((pys - TOP_Y_MIN) // TOP_Y_DIVISION).astype(np.int32)
    # qzs=((pzs-TOP_Z_MIN)//TOP_Z_DIVISION).astype(np.int32)
    qzs = (pzs - TOP_Z_MIN) / TOP_Z_DIVISION
    quantized = np.dstack((qxs, qys, qzs, prs)).squeeze()

    X0, Xn = 0, int((TOP_X_MAX - TOP_X_MIN) // TOP_X_DIVISION) + 1
    Y0, Yn = 0, int((TOP_Y_MAX - TOP_Y_MIN) // TOP_Y_DIVISION) + 1
    Z0, Zn = 0, int((TOP_Z_MAX - TOP_Z_MIN) / TOP_Z_DIVISION)
    height = Xn - X0
    width = Yn - Y0
    channel = Zn - Z0 + 2
    # print('height,width,channel=%d,%d,%d'%(height,width,channel))
    top = np.zeros(shape=(height, width, channel), dtype=np.float32)

    # histogram = Bin(channel, 0, Zn, "z", Bin(height, 0, Yn, "y", Bin(width, 0, Xn, "x", Maximize("intensity"))))
    # histogram.fill.numpy({"x": qxs, "y": qys, "z": qzs, "intensity": prs})

    if 1:  # new method
        for x in range(Xn):
            ix = np.where(quantized[:, 0] == x)
            quantized_x = quantized[ix]
            if len(quantized_x) == 0:
                continue
            yy = -x

            for y in range(Yn):
                iy = np.where(quantized_x[:, 1] == y)
                quantized_xy = quantized_x[iy]
                count = len(quantized_xy)
                if count == 0:
                    continue
                xx = -y

                top[yy, xx, Zn + 1] = min(1, np.log(count + 1) / math.log(32))
                max_height_point = np.argmax(quantized_xy[:, 2])
                top[yy, xx, Zn] = quantized_xy[max_height_point, 3]

                for z in range(Zn):
                    iz = np.where(
                        (quantized_xy[:, 2] >= z) & (quantized_xy[:, 2] <= z + 1)
                    )
                    quantized_xyz = quantized_xy[iz]
                    if len(quantized_xyz) == 0:
                        continue
                    zz = z

                    # height per slice
                    max_height = max(0, np.max(quantized_xyz[:, 2]) - z)
                    top[yy, xx, zz] = max_height

    # if 0: #unprocess
    #     top_image = np.zeros((height,width,3),dtype=np.float32)
    #
    #     num = len(lidar)
    #     for n in range(num):
    #         x,y = qxs[n],qys[n]
    #         if x>=0 and x <width and y>0 and y<height:
    #             top_image[y,x,:] += 1
    #
    #     max_value=np.max(np.log(top_image+0.001))
    #     top_image = top_image/max_value *255
    #     top_image=top_image.astype(dtype=np.uint8)

    return top


MATRIX_Mt = np.array(
    [
        [2.34773698e-04, 1.04494074e-02, 9.99945389e-01, 0.00000000e00],
        [-9.99944155e-01, 1.05653536e-02, 1.24365378e-04, 0.00000000e00],
        [-1.05634778e-02, -9.99889574e-01, 1.04513030e-02, 0.00000000e00],
        [5.93721868e-02, -7.51087914e-02, -2.72132796e-01, 1.00000000e00],
    ]
)

MATRIX_Kt = np.array(
    [[721.5377, 0.0, 0.0], [0.0, 721.5377, 0.0], [609.5593, 172.854, 1.0]]
)


def box3d_to_rgb_box00(box3d):

    # box3d = boxes3d[n]
    Ps = np.hstack((box3d, np.ones((8, 1))))
    Qs = np.matmul(Ps, MATRIX_Mt)
    Qs = Qs[:, 0:3]
    qs = np.matmul(Qs, MATRIX_Kt)
    zs = qs[:, 2].reshape(8, 1)
    qs = qs / zs

    return qs[:, 0:2]


def box3d_to_rgb_box0000(boxes3d, Mt=None, Kt=None):
    # if (cfg.DATA_SETS_TYPE == 'kitti'):
    if Mt is None:
        Mt = np.array(MATRIX_Mt)
    if Kt is None:
        Kt = np.array(MATRIX_Kt)

    num = len(boxes3d)
    projections = np.zeros((num, 8, 2), dtype=np.int32)
    for n in range(num):
        box3d = boxes3d[n]
        Ps = np.hstack((box3d, np.ones((8, 1))))
        Qs = np.matmul(Ps, Mt)
        Qs = Qs[:, 0:3]
        qs = np.matmul(Qs, Kt)
        zs = qs[:, 2].reshape(8, 1)
        qs = qs / zs
        # pts_3d=project_to_image(qs[:,0:2], P)
        projections[n] = qs[:, 0:2]
        # projections[n] = proj3d_to_2d(qs[:,0:2])
        # projections[n] = pts_3d
    return projections


def proj3d_to_2d(rgbpoint):
    x0 = np.min(rgbpoint[:, 0])
    x1 = np.max(rgbpoint[:, 0])
    y0 = np.min(rgbpoint[:, 1])
    y1 = np.max(rgbpoint[:, 1])
    # x0 = max(0,x0)
    # x1 = min(x1, proj.image_width)
    # y0 = max(0,y0)
    # y1 = min(y1, proj.image_height)
    return np.array([x0, y0, x1, y1])


def project_to_image(pts_3d, P):
    """ Project 3d points to image plane.

    Usage: pts_2d = projectToImage(pts_3d, P)
      input: pts_3d: nx3 matrix
             P:      3x4 projection matrix
      output: pts_2d: nx2 matrix

      P(3x4) dot pts_3d_extended(4xn) = projected_pts_2d(3xn)
      => normalize projected_pts_2d(2xn)

      <=> pts_3d_extended(nx4) dot P'(4x3) = projected_pts_2d(nx3)
          => normalize projected_pts_2d(nx2)
    """
    n = pts_3d.shape[0]
    pts_3d_extend = np.hstack((pts_3d, np.ones((n, 1))))
    # print(('pts_3d_extend shape: ', pts_3d_extend.shape))
    pts_2d = np.dot(pts_3d_extend, np.transpose(P))  # nx3
    pts_2d[:, 0] /= pts_2d[:, 2]
    pts_2d[:, 1] /= pts_2d[:, 2]
    return pts_2d[:, 0:2]


def compute_box_3d(obj, P):
    """ Takes an object and a projection matrix (P) and projects the 3d
        bounding box into the image plane.
        Returns:
            corners_2d: (8,2) array in left image coord.
            corners_3d: (8,3) array in in rect camera coord.
    """
    # compute rotational matrix around yaw axis
    R = roty(obj.ry)

    # 3d bounding box dimensions
    l = obj.l
    w = obj.w
    h = obj.h

    # 3d bounding box corners
    x_corners = [l / 2, l / 2, -l / 2, -l / 2, l / 2, l / 2, -l / 2, -l / 2]
    y_corners = [0, 0, 0, 0, -h, -h, -h, -h]
    z_corners = [w / 2, -w / 2, -w / 2, w / 2, w / 2, -w / 2, -w / 2, w / 2]

    # rotate and translate 3d bounding box
    corners_3d = np.dot(R, np.vstack([x_corners, y_corners, z_corners]))
    # print corners_3d.shape
    corners_3d[0, :] = corners_3d[0, :] + obj.t[0]
    corners_3d[1, :] = corners_3d[1, :] + obj.t[1]
    corners_3d[2, :] = corners_3d[2, :] + obj.t[2]
    # print 'cornsers_3d: ', corners_3d
    # only draw 3d bounding box for objs in front of the camera
    if np.any(corners_3d[2, :] < 0.1):
        corners_2d = None
        return corners_2d, np.transpose(corners_3d)

    # project the 3d bounding box into the image plane
    corners_2d = project_to_image(np.transpose(corners_3d), P)
    # print 'corners_2d: ', corners_2d
    return corners_2d, np.transpose(corners_3d)


def compute_orientation_3d(obj, P):
    """ Takes an object and a projection matrix (P) and projects the 3d
        object orientation vector into the image plane.
        Returns:
            orientation_2d: (2,2) array in left image coord.
            orientation_3d: (2,3) array in in rect camera coord.
    """

    # compute rotational matrix around yaw axis
    R = roty(obj.ry)

    # orientation in object coordinate system
    orientation_3d = np.array([[0.0, obj.l], [0, 0], [0, 0]])

    # rotate and translate in camera coordinate system, project in image
    orientation_3d = np.dot(R, orientation_3d)
    orientation_3d[0, :] = orientation_3d[0, :] + obj.t[0]
    orientation_3d[1, :] = orientation_3d[1, :] + obj.t[1]
    orientation_3d[2, :] = orientation_3d[2, :] + obj.t[2]

    # vector behind image plane?
    if np.any(orientation_3d[2, :] < 0.1):
        orientation_2d = None
        return orientation_2d, np.transpose(orientation_3d)

    # project orientation into the image plane
    orientation_2d = project_to_image(np.transpose(orientation_3d), P)
    return orientation_2d, np.transpose(orientation_3d)


def draw_projected_box3d(image, qs, color=(0, 255, 0), thickness=2):
    """ Draw 3d bounding box in image
        qs: (8,3) array of vertices for the 3d box in following order:
            1 -------- 0
           /|         /|
          2 -------- 3 .
          | |        | |
          . 5 -------- 4
          |/         |/
          6 -------- 7
    """
    qs = qs.astype(np.int32)
    for k in range(0, 4):
        # Ref: http://docs.enthought.com/mayavi/mayavi/auto/mlab_helper_functions.html
        i, j = k, (k + 1) % 4
        # use LINE_AA for opencv3
        # cv2.line(image, (qs[i,0],qs[i,1]), (qs[j,0],qs[j,1]), color, thickness, cv2.CV_AA)
        cv2.line(image, (qs[i, 0], qs[i, 1]), (qs[j, 0], qs[j, 1]), color, thickness)
        i, j = k + 4, (k + 1) % 4 + 4
        cv2.line(image, (qs[i, 0], qs[i, 1]), (qs[j, 0], qs[j, 1]), color, thickness)

        i, j = k, k + 4
        cv2.line(image, (qs[i, 0], qs[i, 1]), (qs[j, 0], qs[j, 1]), color, thickness)
    return image


def draw_top_image(lidar_top):
    top_image = np.sum(lidar_top, axis=2)
    top_image = top_image - np.min(top_image)
    divisor = np.max(top_image) - np.min(top_image)
    top_image = top_image / divisor * 255
    top_image = np.dstack((top_image, top_image, top_image)).astype(np.uint8)
    return top_image


def draw_box3d_on_top(
    image,
    boxes3d,
    color=(255, 255, 255),
    thickness=1,
    scores=None,
    text_lables=[],
    is_gt=False,
):

    # if scores is not None and scores.shape[0] >0:
    # print(scores.shape)
    # scores=scores[:,0]
    font = cv2.FONT_HERSHEY_SIMPLEX
    img = image.copy()
    num = len(boxes3d)
    startx = 5
    for n in range(num):
        b = boxes3d[n]
        x0 = b[0, 0]
        y0 = b[0, 1]
        x1 = b[1, 0]
        y1 = b[1, 1]
        x2 = b[2, 0]
        y2 = b[2, 1]
        x3 = b[3, 0]
        y3 = b[3, 1]
        u0, v0 = lidar_to_top_coords(x0, y0)
        u1, v1 = lidar_to_top_coords(x1, y1)
        u2, v2 = lidar_to_top_coords(x2, y2)
        u3, v3 = lidar_to_top_coords(x3, y3)
        if is_gt:
            color = (0, 255, 0)
            startx = 5
        else:
            color = heat_map_rgb(0.0, 1.0, scores[n]) if scores is not None else 255
            startx = 85
        cv2.line(img, (u0, v0), (u1, v1), color, thickness, cv2.LINE_AA)
        cv2.line(img, (u1, v1), (u2, v2), color, thickness, cv2.LINE_AA)
        cv2.line(img, (u2, v2), (u3, v3), color, thickness, cv2.LINE_AA)
        cv2.line(img, (u3, v3), (u0, v0), color, thickness, cv2.LINE_AA)
    for n in range(len(text_lables)):
        text_pos = (startx, 25 * (n + 1))
        cv2.putText(img, text_lables[n], text_pos, font, 0.5, color, 0, cv2.LINE_AA)
    return img


# hypothesis function
def hypothesis_func(w, x):
    w1, w0 = w
    return w1 * x + w0


# error function
def error_func(w, train_x, train_y):
    return hypothesis_func(w, train_x) - train_y


def dump_fit_func(w_fit):
    w1, w0 = w_fit
    print("fitting line=", str(w1) + "*x + " + str(w0))
    return


# square error
def dump_fit_cost(w_fit, train_x, train_y):
    error = error_func(w_fit, train_x, train_y)
    square_error = sum(e * e for e in error)
    print("fitting cost:", str(square_error))
    return square_error


def linear_regression(train_x, train_y, test_x):
    # train set
    # train_x = np.array([8.19,2.72,6.39,8.71,4.7,2.66,3.78])
    # train_y = np.array([7.01,2.78,6.47,6.71,4.1,4.23,4.05])

    # linear regression by leastsq
    # msg = "invoke scipy leastsq"
    w_init = [20, 1]  # weight factor init
    fit_ret = leastsq(error_func, w_init, args=(train_x, train_y))
    w_fit = fit_ret[0]

    # dump fit result
    dump_fit_func(w_fit)
    dump_fit_cost(w_fit, train_x, train_y)

    # test set
    # test_x = np.array(np.arange(train_x.min(), train_x.max(), 1.0))
    test_y = hypothesis_func(w_fit, test_x)
    test_y0 = hypothesis_func(w_fit, train_x)
    return test_y, test_y0



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/17 13:02:51

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from collections import defaultdict
from io import StringIO
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['g_factory']

class ObjectFactory:
    def __init__(self):
        self._creator_list = defaultdict(list)
        self._name_list = defaultdict(list)

    def get_creator(self, _name, category=None, *args, **kws):
        for creator in self.get_creator_list(category):
            _ctor = creator(_name, *args, **kws)
            if _ctor:
                return _ctor

    def get_instance(self, _name, category=None, _args=None, _kws=None, *args, **kws):
        if _args is None: _args = ()
        if _kws is None: _kws = {}
        if args is None: args = ()
        ctor = self.get_creator(category, _name, *_args, **_kws)
        if ctor:
            return ctor(*args, **kws)

    def _add_name(self, _list, name):
        if name:
            _list.append(name)

    def add_name(self, category, *names):
        _list = self._name_list[category]
        for name in names:
            if isinstance(name, (tuple, list)):
                for _name in name:
                    self._add_name(_list, _name)
            elif name:
                self._add_name(_list, name)

    def add_creator(self, category, creator):
        self._creator_list[category].append(creator)

    def register(self, category, *names):
        self.add_name(category, *names)
        def inner(creator):
            self.add_creator(category, creator)
            def inner2(name):
                return creator(name)
            return inner2
        return inner

    def get_name_list(self, category=None):
        if category:
            return self._name_list[category]
        else:
            _name_list = []
            for v in self._name_list.values():
                _name_list.extend(v)
            return _name_list

    def get_creator_list(self, category=None):
        if category:
            return self._creator_list[category]
        else:
            _ctor_list = []
            for v in self._creator_list.values():
                 _ctor_list.extend(v)
            return _ctor_list

    def __str__(self):
        s = StringIO()
        width = 80
        for cat, namelist in self._name_list.items():
            s.write(cat + f" ({len(namelist)})\n")
            _w, _i, _len = 0, 0, len(namelist)
            for i in range(0, _len):
                _w += len(namelist[i])
                if _w > width:
                    s.write("\t" + ", ".join(namelist[_i:i-1]) + "\n")
                    _i, _w = i-1, 0
            s.write("\t" + ", ".join(namelist[_i:_len]) + "\n")
        return s.getvalue()

# 全局注册工厂
g_factory = ObjectFactory()


from .bin import *
from .contours import *
from .cvt import *
from .draw import *
from .edge import *
from .filter import *
from .geo import *
from .info import *
from .io import *
from .morph import *
from .resize import *
from .show import *
from .util import *


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/8/31 20:14:34

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import sys, types
import inspect as inspect
from pprint import pprint
# ===============================================================================
r'''提供一些调试方便的函数
'''
# ===============================================================================
__all__ = ['pprint',
           'print_args','print_eval',
           'print_type', 'print_repr', 'print_signature',
           ]

def print_args(*args, **kws):
    print('args:', args)
    print('kws:', kws)

def print_eval(st):
    '''字符串打印并执行'''
    _frm = sys._getframe(1)
    val = eval(st, _frm.f_globals, _frm.f_locals)
    print( f"{st} = {val}")

def print_type(t):
    '''打印类别'''
    _type_name = None
    for tname in dir(types):
        if tname[0].isupper() and isinstance(t, getattr(types, tname)):
            _type_name = tname
            break
    if _type_name is None:
        _type_name = f"{type(t)}"
    print(_type_name)

def print_repr(v):
    '''打印 repr '''
    print(repr(v))

def print_signature(func):
    sig = inspect.signature(func)
    print(f'{func.__name__}{str(sig)}')




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/8/13 10:14:58

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from collections import OrderedDict
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['load_csv', 'save_csv']

def load_csv(fname):
    ret = OrderedDict()
    with open(fname, mode='r', encoding='utf8') as f:
        fields = f.readline().strip().split(',')
        for field in fields:
            ret[field] = []

        for line in f:
            for i,seg in enumerate(line.strip().split(',')):
                ret[fields[i]].append(seg)
    return ret

def save_csv(fname):
    pass


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/7/21 10:33:17

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os, sys
from PyQt5.QtWidgets import QWidget, QMainWindow, QDialog
from PyQt5.uic import loadUi, compileUi
from laok.base import log_info
#===============================================================================
#
#===============================================================================

__all__ = ['UiMainWindowDy', 'UiDyWidget', 'UiDyDialog', 'widget_from_ui', 'compile_ui_file']

def _init_ui_file(obj):
    if hasattr(obj, 'ui_file'):
        ui_file = getattr(obj, 'ui_file')
    else:
        module_file = sys.modules[obj.__module__].__file__
        ui_file = os.path.splitext(module_file)[0] + ".ui"
    if not os.path.exists(ui_file):
        raise ValueError(f"you need config 'ui_file' in your class of[{obj.__class__}]")
    else:
        log_info(f' load ui file= {ui_file}')
    loadUi(ui_file, obj)  # 初始化Ui

def widget_from_ui(ui_file):
    log_info(f' load ui file= {ui_file}')
    return loadUi(ui_file)

class UiMainWindowDy(QMainWindow):
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        QMainWindow.__init__(obj)
        _init_ui_file(obj)
        return obj

    def __init__(self):
        pass

class UiDyWidget(QWidget):
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        QWidget.__init__(obj)
        _init_ui_file(obj)
        return obj

    def __init__(self):
        pass

class UiDyDialog(QDialog):
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        QDialog.__init__(obj)
        _init_ui_file(obj)
        return obj

    def __init__(self):
        pass

def compile_ui_file(ui_file, save_file=None):
    '''
        保存 X_ui.py 文件
    '''
    if save_file is None:
        save_file = os.path.splitext(ui_file)[0] + '_ui.py'
    with open(save_file, 'w', encoding='utf8') as f:
        compileUi(uifile=ui_file, pyfile=f)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/13 21:23:49

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
from .op import mat_angle_axis
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['CldScale', 'CldTranslate', 'CldNormal', 'CldTransform', 'CldRotateX', 'CldRotateY', 'CldRotateZ',
           'CldRotateXYZ']

class CldScale:
    def __init__(self, scale_x = 1.0, scale_y = 1.0, scale_z = 1.0):
        self.x = scale_x
        self.y = scale_y
        self.z = scale_z

    def __call__(self, points):
        points[:, 0] *= self.x
        points[:, 1] *= self.y
        points[:, 2] *= self.z
        return points

class CldTranslate(object):
    def __init__(self, shift_x = 0, shift_y = 0, shift_z = 0):
        self.x = shift_x
        self.y = shift_y
        self.z = shift_z

    def __call__(self, points):
        points[:, 0] += self.x
        points[:, 1] += self.y
        points[:, 2] += self.z
        return points

class CldTransform:
    def __init__(self, mat):
        self.mat = mat

    def __call__(self, points):
        has_normal = points.shape[1] > 3
        if not has_normal:
            return np.matmul(points, np.transpose(self.mat))
        else:
            pc_xyz = points[:, 0:3]
            pc_normals = points[:, 3:]
            points[:, 0:3] = np.matmul(pc_xyz, np.transpose(self.mat))
            points[:, 3:] = np.matmul(pc_normals, np.transpose(self.mat))
            return points

class CldRotateX:
    def __init__(self, angle = 90):
        self.angle = angle

    def __call__(self, points):
        axis = np.array([1.0, 0.0, 0.0])
        mat = mat_angle_axis(self.angle, axis)
        return CldTransform(mat)(points)

class CldRotateY:
    def __init__(self, angle = 90):
        self.angle = angle

    def __call__(self, points):
        axis = np.array([0.0, 1.0, 0.0])
        mat = mat_angle_axis(self.angle, axis)
        return CldTransform(mat)(points)

class CldRotateZ:
    def __init__(self, angle = 90):
        self.angle = angle

    def __call__(self, points):
        axis = np.array([0.0, 0.0, 1.0])
        mat = mat_angle_axis(self.angle, axis)
        return CldTransform(mat)(points)


class CldRotateXYZ:
    def __init__(self, angle_x=90, angle_y=0, angel_z=0):
        self.angle_x = angle_x
        self.angle_y = angle_y
        self.angle_z = angel_z

    def __call__(self, points):
        Rx = mat_angle_axis(self.angle_x, np.array([1.0, 0.0, 0.0]))
        Ry = mat_angle_axis(self.angle_y, np.array([0.0, 1.0, 0.0]))
        Rz = mat_angle_axis(self.angle_z, np.array([0.0, 0.0, 1.0]))
        mat = np.matmul(np.matmul(Rz, Ry), Rx)
        return CldTransform(mat)(points)

class CldNormal:
    def __call__(self, points):
        centroid = np.mean(points, axis=0)
        points -= centroid
        m = np.max(np.sqrt(np.sum(points ** 2, axis=1)))
        points /= m
        return points


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/6/4 17:25:47

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import inspect, sys, time, traceback, math, re, codecs
from datetime import datetime
#===============================================================================
# 交互式运行
#===============================================================================
PRINT_FUNC_COUNT = 3

__all__ = ['run', 'main_module']
main_module = sys.modules.get('__main__')

# 执行计时
def _time_run(_kvfunc, _repeat = 1, **kws):
    name, func = _kvfunc
    print('{}==begin[date:{}]'.format(name, datetime.now().strftime('%Y-%m-%d %H:%M:%S')))

    t0 = time.time()
    for i in range(_repeat):
        try:
            func(**kws)
        except (SystemExit, KeyboardInterrupt):
            break
        except Exception as e:
            print ( '\nexception in call function---->\n%s' % traceback.format_exc() )
            break
    elapse = time.time() - t0
    print ( '{}==end[use time:{}s]\n'.format(name, elapse) )

#从文件提取特定函数名字,主要是用于排序
def _find_order_names(filename, re_pat):
    with codecs.open(filename, encoding='utf8') as f:
        for i,line in enumerate(f, 1):
            res = re_pat.match(line)
            if res:
                yield i,res.group(1)

# 删除头尾
def _trim(name, _suffix):
    if name.startswith(_suffix):
        name = name[len(_suffix):]
    if name.endswith(_suffix):
        name = name[:-len(_suffix)]
    return name

def run(_opt='last', _suffix ='_lk', _module = None, _repeat = 1, **kws):
    '''执行以 _suffix 结尾的函数
    :param _opt: last/first/select/all/function_name
    :return:
    '''
    RE_PAT = re.compile(r"^def\s*(.*?{pat})|({pat}.*?)\(.*?\):.*".format(pat = _suffix))

    # 获取所有 _suffix 结尾的函数
    if _module is None:
        _module = main_module
    kvList = [(_trim(k, _suffix), v)
              for k,v in inspect.getmembers(_module, inspect.isfunction)
              if k.endswith(_suffix)]

    # 检测是否有测试用例
    if not kvList:
        return

    # 按照代码文本顺序排序
    _kvOrder = {}
    for order, name in  _find_order_names(_module.__file__, RE_PAT):
        _kvOrder[_trim(name, _suffix)] = order
    kvList.sort(key = lambda pair: _kvOrder[pair[0]])

    # 根据选择,执行函数
    if _opt == 'first':
        return _time_run(kvList[0], _repeat, **kws)
    elif _opt == 'last':
        return _time_run(kvList[-1], _repeat,**kws)
    elif _opt == 'all':
        for kv in kvList:
            _time_run(kv, _repeat, **kws)
    elif _opt in ['select','']:
        _len = len(kvList)
        idx_w = math.ceil(math.log10(_len))  #索引的宽度
        idx_name_w = {} #名字列的宽度
        for i, (k,_) in enumerate(kvList):
            _kw = len(k)
            i = i % PRINT_FUNC_COUNT
            if i in idx_name_w:
                idx_name_w[i] = max( _kw, idx_name_w[i] )
            else:
                idx_name_w[i] = _kw

        while True: # 循环执行任务
            for i, (k,_) in enumerate(kvList):
                fmt = "<%0{}s>%-{}s".format(idx_w, idx_name_w[ i % PRINT_FUNC_COUNT])
                # print(fmt, end=' ')
                print( fmt % (i, k), end=" ")
                if (i+1) % PRINT_FUNC_COUNT == 0:
                    print()
            if (i+1) % PRINT_FUNC_COUNT != 0:
                print()

            try:
                i = input('which func do you want to run , -1=quit , -2=run-all:')
                i = int(i) # 输入索引
            except:
                i = _len - 1 #执行最后一个

            if i == -2: #运行所有的task,退出
                for kv in kvList:
                    _time_run(kv, _repeat, **kws)
                return
            elif i == -1: #直接退出
                return
            else: #执行选择的任务
                i = (i + _len) % _len
                _time_run( kvList[i], _repeat,**kws)

            if _opt == '': #执行一次,则退出
                return
    else : #否则判断是否执行任务列表
        if not isinstance(_opt, (list,tuple) ):
            run_list = (_opt, )
        else:
            run_list = _opt

        for _run_op in run_list:
            for kv in kvList:
                if (isinstance(_run_op, str) and kv[0] == _trim(_run_op, _suffix) ) \
                        or (inspect.isfunction(_run_op) and kv[1] == _run_op):
                    _time_run(kv, _repeat,**kws)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/17 21:25:44

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import laok
from tqdm import tqdm
from ..op.device import get_device
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['infer_cls_one', 'infer_cls_batch', 'infer_cls_dataloader']

def infer_cls_one(model, data):
    batch_data = data.unsqueeze(0)
    pred_choice = infer_cls_batch(model, batch_data)
    return pred_choice[0]

def infer_cls_batch(model, data):
    model.eval()
    _device = get_device(model)
    with torch.no_grad():
        pred = model(data.to(_device))
        pred_choice = pred.max(1)[1]
        return [v.item() for v in pred_choice]

def infer_cls_dataloader(model, loader):
    mean_correct = []
    # class_acc = None
    laok.log_info('val data_loader')

    pbar = tqdm(enumerate(loader), total=len(loader), smoothing=0.9)
    for j, (data, target) in pbar:
        d_size_str = laok.size_to_str(data.size())
        t_size_str = laok.size_to_str(target.size())
        pbar.set_description(f"validate data:{d_size_str} target:{t_size_str}")
        mean_val, pred_choice = infer_cls_batch(model, data, target)
        mean_correct.append(mean_val)
        # print(f"mean_val:{mean_val}")
    #     if num_classes:
    #         if class_acc is None:
    #             class_acc = np.zeros((num_classes, 3))
    #         for cat in np.unique(target.cpu()):
    #             classacc = pred_choice.cpu()[target == cat].eq(target[target == cat].long()).cpu().sum()
    #
    #             val = float(data[target == cat].size()[0])
    #             if val > 0:
    #                 class_acc[cat, 0] += classacc.item() / val
    #                 class_acc[cat, 1] += 1
    #
    # instance_acc = np.mean(mean_correct)
    # if num_classes:
    #     class_acc[:, 2] = class_acc[:, 0] / class_acc[:, 1]
    #     class_acc = np.mean(class_acc[:, 2])
    #     laok.log_info(f"validate instance_acc:{instance_acc}, class_acc:{class_acc}")
    # else:
    #     laok.log_info(f"validate instance_acc:{instance_acc}")

    # return instance_acc, class_acc







#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/3 12:54:29

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from laok.base.fs import file_read_chunk, file_bin_writer, file_read_bin, file_write_bin
from Crypto.Cipher import AES
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['aes_decrypt_cbc_data', 'aes_decrypt_cbc_file', 'aes_decrypt_cbc_file_']

def aes_decrypt_cbc_data(data, key, iv=b'0000000000000000', **kws):
    aes = AES.new(key=key, mode=AES.MODE_CBC, iv=iv, **kws)
    return aes.decrypt(data)

def aes_decrypt_cbc_file(infile, outfile, key, iv=b'0000000000000000', **kws):
    with file_bin_writer(outfile) as f:
        aes = AES.new(key=key, mode=AES.MODE_CBC, iv=iv, **kws)
        for data in file_read_chunk(infile):
            data = aes.decrypt(data)
            f.write(data)

def aes_decrypt_cbc_file_(fname, key, iv=b'0000000000000000', **kws):
    data = file_read_bin(fname)
    data = aes_decrypt_cbc_data(data, key, iv, **kws)
    file_write_bin(fname, data)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/8 20:03:21

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
from ..trans.op import mat_angle_axis
from laok.base.alg import arr_split_n
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ =[
          'CldScaleRand', 'CldTranslateRand', 'CldJitter', 'CldDropout',
          'CldRotateRand', 'CldRotatePerturbationRand',
          'CldRepeat'
          ]

class CldScaleRand:
    def __init__(self, lo=0.8, hi=1.25):
        self.lo, self.hi = lo, hi

    def __call__(self, points):
        scaler = np.random.uniform(self.lo, self.hi)
        points[:] *= scaler
        return points

class CldTranslateRand:
    def __init__(self, shift_range=0.1):
        self.shift_range = shift_range

    def __call__(self, points):
        translation = np.random.uniform(-self.shift_range, self.shift_range)
        points[:] += translation
        return points

class CldRotateRand:
    def __init__(self, axis=np.array([0.0, 1.0, 0.0])):
        self.axis = axis

    def __call__(self, points):
        rotation_angle = np.random.uniform() * 2 * np.pi
        rm = mat_angle_axis(rotation_angle, self.axis)

        normals = points.shape[1] > 3
        if not normals:
            return np.matmul(points, np.transpose(rm))
        else:
            pc_xyz = points[:, 0:3]
            pc_normals = points[:, 3:]
            points[:, 0:3] = np.matmul(pc_xyz, np.transpose(rm))
            points[:, 3:] = np.matmul(pc_normals, np.transpose(rm))
            return points

class CldRotatePerturbationRand:
    def __init__(self, angle_sigma=0.06, angle_clip=0.18):
        self.angle_sigma, self.angle_clip = angle_sigma, angle_clip

    def _get_angles(self):
        angles = np.clip(
            self.angle_sigma * np.random.randn(3), -self.angle_clip, self.angle_clip
        )
        return angles

    def __call__(self, points):
        angles = self._get_angles()
        Rx = mat_angle_axis(angles[0], np.array([1.0, 0.0, 0.0]))
        Ry = mat_angle_axis(angles[1], np.array([0.0, 1.0, 0.0]))
        Rz = mat_angle_axis(angles[2], np.array([0.0, 0.0, 1.0]))

        rm = np.matmul(np.matmul(Rz, Ry), Rx)

        normals = points.shape[1] > 3
        if not normals:
            return np.matmul(points, np.transpose(rm))
        else:
            pc_xyz = points[:, 0:3]
            pc_normals = points[:, 3:]
            points[:, 0:3] = np.matmul(pc_xyz, np.transpose(rm))
            points[:, 3:] = np.matmul(pc_normals, np.transpose(rm))
            return points

class CldJitter(object):
    def __init__(self, std=0.01, clip=0.05):
        self.std, self.clip = std, clip

    def __call__(self, points):
        N = points.shape[0]
        jittered_data = np.clip(self.std * np.random.randn(N, 3), -1*self.clip, self.clip)
        points[:, 0:3] += jittered_data
        return points

class CldDropout(object):
    def __init__(self, max_dropout_ratio=0.875):
        assert max_dropout_ratio >= 0 and max_dropout_ratio < 1
        self.max_dropout_ratio = max_dropout_ratio

    def __call__(self, points):
        dropout_ratio = np.random.random() * self.max_dropout_ratio  # 0~0.875
        drop_idx = np.where(np.random.random((points.shape[0])) <= dropout_ratio)[0]
        if len(drop_idx) > 0:
            points[drop_idx] = points[0]  # set to the first point
        return points  # torch.from_numpy(pc).float()

class CldRepeat:
    def __init__(self, count=4):
        self.count = count

    def __call__(self, points):
        sub_points = arr_split_n(points, self.count)

        points1, points2 = arr_split_n(points, self.count)
        x1 = np.ptp(points1, axis=0)[0]
        x2 = np.ptp(points2, axis=0)[0]
        points1[:] -= np.array([x1/2+0.01, 0, 0])
        points2[:] += np.array([x2/2+0.01, 0, 0])
        return np.concatenate(sub_points, axis=0)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/21 17:01:47
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import math
import numpy as np
from scipy.spatial.transform import Rotation
#===============================================================================
r'''
'''
#===============================================================================
__all__ = [
            'CldRotate',
            'CldRotatePerturbation',
            'rotate_matrix_from_angle_axis',
            'is_rotation_matrix',
            'rotation_matrix_to_euler',
            'euler_to_rotate_matrix',
            'euler_to_rotate_vec',
           ]



class CldRotate(object):
    def __init__(self, axis=np.array([0.0, 1.0, 0.0])):
        self.axis = axis

    def __call__(self, points):
        rotation_angle = np.random.uniform() * 2 * np.pi
        rotation_matrix = rotate_matrix_from_angle_axis(rotation_angle, self.axis)

        normals = points.shape[1] > 3
        if not normals:
            return np.matmul(points, rotation_matrix.transpose())
        else:
            pc_xyz = points[:, 0:3]
            pc_normals = points[:, 3:]
            points[:, 0:3] = np.matmul(pc_xyz, rotation_matrix.transpose())
            points[:, 3:] = np.matmul(pc_normals, rotation_matrix.transpose())

            return points


class CldRotatePerturbation(object):
    def __init__(self, angle_sigma=0.06, angle_clip=0.18):
        self.angle_sigma, self.angle_clip = angle_sigma, angle_clip

    def _get_angles(self):
        angles = np.clip(
            self.angle_sigma * np.random.randn(3), -self.angle_clip, self.angle_clip
        )

        return angles

    def __call__(self, points):
        angles = self._get_angles()
        Rx = rotate_matrix_from_angle_axis(angles[0], np.array([1.0, 0.0, 0.0]))
        Ry = rotate_matrix_from_angle_axis(angles[1], np.array([0.0, 1.0, 0.0]))
        Rz = rotate_matrix_from_angle_axis(angles[2], np.array([0.0, 0.0, 1.0]))

        rotation_matrix = np.matmul(np.matmul(Rz, Ry), Rx)

        normals = points.shape[1] > 3
        if not normals:
            return np.matmul(points, rotation_matrix.transpose())
        else:
            pc_xyz = points[:, 0:3]
            pc_normals = points[:, 3:]
            points[:, 0:3] = np.matmul(pc_xyz, rotation_matrix.transpose())
            points[:, 3:] = np.matmul(pc_normals, rotation_matrix.transpose())

            return points

def rotate_matrix_from_angle_axis(angle, axis):
    # type: (float, np.ndarray) -> float
    r"""Returns a 3x3 rotation matrix that performs a rotation around axis by angle

    Parameters
    ----------
    angle : float
        Angle to rotate by
    axis: np.ndarray
        Axis to rotate about

    Returns
    -------
    np.ndarray
        3x3 rotation matrix
    """
    u = axis / np.linalg.norm(axis)
    cosval, sinval = np.cos(angle), np.sin(angle)

    # yapf: disable
    cross_prod_mat = np.array([[0.0, -u[2], u[1]],
                                [u[2], 0.0, -u[0]],
                                [-u[1], u[0], 0.0]])

    R = cosval * np.eye(3) \
        + sinval * cross_prod_mat \
        + (1.0 - cosval) * np.outer(u, u)
    return R


def is_rotation_matrix(R) :
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype = R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6


def rotation_matrix_to_euler(R, err=1e-6) :
    sy = math.sqrt(R[0,0] * R[0,0] +  R[1,0] * R[1,0])
    singular = sy < err
    if  not singular :
        x = math.atan2(R[2,1] , R[2,2])
        y = math.atan2(-R[2,0], sy)
        z = math.atan2(R[1,0], R[0,0])
    else :
        x = math.atan2(-R[1,2], R[1,1])
        y = math.atan2(-R[2,0], sy)
        z = 0
    return np.array([x, y, z])

def euler_to_rotate_matrix(yaw, pitch, roll):
    Rz_yaw = np.array([
        [np.cos(yaw), -np.sin(yaw), 0],
        [np.sin(yaw), np.cos(yaw), 0],
        [0, 0, 1]])
    Ry_pitch = np.array([
        [np.cos(pitch), 0, np.sin(pitch)],
        [0, 1, 0],
        [-np.sin(pitch), 0, np.cos(pitch)]])
    Rx_roll = np.array([
        [1, 0, 0],
        [0, np.cos(roll), -np.sin(roll)],
        [0, np.sin(roll), np.cos(roll)]])
    # R = RzRyRx
    rotMat = np.dot(Rz_yaw, np.dot(Ry_pitch, Rx_roll))
    return rotMat

def euler_to_rotate_vec(yaw, pitch, roll):
    # compute the rotation matrix
    Rmat = euler_to_rotate_matrix(yaw, pitch, roll)

    theta = math.acos(((Rmat[0, 0] + Rmat[1, 1] + Rmat[2, 2]) - 1) / 2)
    sin_theta = math.sin(theta)
    if sin_theta == 0:
        rx, ry, rz = 0.0, 0.0, 0.0
    else:
        multi = 1 / (2 * math.sin(theta))
        rx = multi * (Rmat[2, 1] - Rmat[1, 2]) * theta
        ry = multi * (Rmat[0, 2] - Rmat[2, 0]) * theta
        rz = multi * (Rmat[1, 0] - Rmat[0, 1]) * theta
    return rx, ry, rz




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/9/18 11:31:01

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from .._creator import ctor
# ===============================================================================
r'''
'''
# ===============================================================================

name_list = ['Modelnet',
             'ModelnetIneratia',
             'ModelnetNormals',
             'ModelnetNormalsDiff']

@ctor.register(*name_list)
def _get(_name, **kws):
    DS = None

    if _name == 'Modelnet':
        from .model_net import Modelnet as DS
    elif _name == 'ModelnetIneratia':
        from .model_net_ineratia import ModelnetIneratia as DS
    elif _name == 'ModelnetNormals':
        from .model_net_normals import ModelnetNormals as DS
    elif _name == 'ModelnetNormalsDiff':
        from .model_net_normals import ModelnetDiffNormals as DS

    if DS:
        return DS(**kws)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/6/7 00:23:50

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import urllib.parse as _url_parse
from ..fs import path_join
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['url_join',  'url_split', 'url_make_file']

def url_join(base, url, allow_fragments=True):
    return _url_parse.urljoin(base, url, allow_fragments=allow_fragments)

def url_split(url, scheme='', allow_fragments=True):
    '''http://user:pwd@NetLoc:80/p1;para/p2;para?query=arg#frag
        scheme  : http
        netloc  : user:pwd@NetLoc:80
        path    : /p1;para/p2;para
        query   : query=arg
        fragment: frag
        username: user
        password: pwd
        hostname: netloc
        port    : 80
    '''
    return _url_parse.urlsplit(url, scheme, allow_fragments=allow_fragments)

def url_make_file(url):
    res = url_split(url)
    netloc = res.netloc
    path = res.path[1:]
    netloc = netloc.replace(':', '_')
    fname = path_join(netloc, path, make_parent=True)
    return fname



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/13 11:45:36

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
from laok import kws_merge
import torch.nn as nn
# ===============================================================================
r'''
Conv → Batch Norm → Activate Function
    这是最为常见的组合顺序，广泛应用于众多经典网络架构，像 ResNet 就采用了这种方式。
    原理剖析：卷积操作之后，运用批量归一化对数据分布加以调整，接着再通过激活函数引入非线性因素。
    优势所在：能有效缓解梯度消失问题，使网络的训练过程更加稳定，还可加快模型的收敛速度。
    适用场景：适合大多数的深度学习任务，特别是在图像分类任务中表现出色。
'''
# ===============================================================================
__all__ = ['Conv1BatchRelu']

class _ConvNormActivate(nn.Module):
    def __init__(self, conv_cls=None, conv_kws=None, norm_cls=None, norm_kws=None, act_cls=None, act_kws=None):
        super().__init__()
        self.conv = conv_cls(**kws_merge(conv_kws))
        self.bn = norm_cls(**kws_merge(norm_kws))
        self.activation = act_cls(**kws_merge(act_kws))

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.activation(x)
        return x

class Conv1BatchRelu(_ConvNormActivate):
    def __init__(self, in_channels, out_channels, kernel_size=1, conv_kws=None, norm_kws=None, act_kws=None):
        conv_kws = kws_merge(conv_kws, in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size)
        norm_kws = kws_merge(norm_kws, num_features=out_channels)
        act_kws = kws_merge(act_kws, inplace=True)
        super().__init__(conv_cls = nn.Conv1d, norm_cls = nn.BatchNorm1d, act_cls = nn.ReLU,
                         conv_kws = conv_kws, norm_kws = norm_kws, act_kws = act_kws)






















#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/17 19:56:46

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import yaml
from collections import OrderedDict
from laok.base.alg import Dict
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['save_yaml_file', 'load_yaml_file']

def save_yaml_file(filename, obj, indent=2, **kws):
    with open(filename, mode='w', encoding='utf8') as f:
        if isinstance(obj, (OrderedDict, Dict) ):
            _dict_dump(obj, f, indent=indent, **kws)
        else:
            yaml.safe_dump(data=obj, stream=f, indent=indent,**kws)
    return True

def load_yaml_file(filename, **kws):
    with open(filename, encoding='utf8') as f:
        return yaml.safe_load(f, **kws)

def _dict_dump(data, stream=None, **kwds):
    class DictDumper(yaml.Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items())

    DictDumper.add_representer(OrderedDict, _dict_representer)
    DictDumper.add_representer(Dict, _dict_representer)
    return yaml.dump(data, stream, DictDumper, **kwds)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/9/16 15:27:22

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import numpy as np
import laok.util.fs as kpath
import laok.cv2d.io as kcv2
from laok.cv3d import inverse_rigid_trans
import laok.util.log as klog
from . import utils
# ===============================================================================
r'''KITTI LiDar 点云数据集
'''
# ===============================================================================

class kitti_object(object):
    """主要支持
    training
        velodyne
        depth
        pred
        image_2
        label_2
        calib
    testing
    """
    def __init__(self, root, split="training", show_info=False):
        self.root = root
        self.split = split

        klog.debug(f'===== {self.__class__.__name__}')
        klog.debug(f'root={root}')
        klog.debug(f'split={split}, show_info={show_info}')

        self.split_dir = os.path.join(root, split)
        self.image_dir = os.path.join(self.split_dir, "image_2")
        self.label_dir = os.path.join(self.split_dir, "label_2")
        self.calib_dir = os.path.join(self.split_dir, "calib")

        self.depthpc_dir = os.path.join(self.split_dir, "depth_pc")
        self.lidar_dir = os.path.join(self.split_dir, "velodyne")
        self.depth_dir = os.path.join(self.split_dir, "depth")
        self.pred_dir = os.path.join(self.split_dir, "pred")

        self._file_list = list(kpath.files_under(self.split_dir, '.bin'))

    def __len__(self):
        return len(self._file_list)

    def get_image(self, idx):
        img_filename = os.path.join(self.image_dir, "%06d.png" % (idx))
        return kcv2.read_color_cv(img_filename)

    def get_lidar(self, idx):
        lidar_filename = os.path.join(self.lidar_dir, "%06d.bin" % (idx))
        scan = np.fromfile(lidar_filename, dtype=np.float32)
        return scan.reshape((-1, 4))

    def get_calibration(self, idx):
        calib_filename = os.path.join(self.calib_dir, "%06d.txt" % (idx))
        return utils.Calibration(calib_filename)

    def get_label_objects(self, idx):
        label_filename = os.path.join(self.label_dir, "%06d.txt" % (idx))

        object_list = []
        with open(label_filename, 'r', encoding='utf8') as f:
            for line in f:
                object_list.append(Object3d(line))
        return object_list


class Object3d(object):
    """ 3d object label """

    def __init__(self, label_file_line):
        data = label_file_line.split(" ")
        data[1:] = [float(x) for x in data[1:]]

        # extract label, truncation, occlusion
        self.type = data[0]  # 'Car', 'Pedestrian', ...
        self.truncation = data[1]  # truncated pixel ratio [0..1]
        self.occlusion = int(data[2] )  # 0=visible, 1=partly occluded, 2=fully occluded, 3=unknown
        self.alpha = data[3]  # object observation angle [-pi..pi]

        # extract 2d bounding box in 0-based coordinates
        self.xmin = data[4]  # left
        self.ymin = data[5]  # top
        self.xmax = data[6]  # right
        self.ymax = data[7]  # bottom
        self.box2d = np.array([self.xmin, self.ymin, self.xmax, self.ymax])

        # extract 3d bounding box information
        self.height = data[8]  # box height
        self.width = data[9]  # box width
        self.length = data[10]  # box length (in meters)
        self.dimensions = (data[8], data[9], data[10])
        self.pos_x = data[11]
        self.pos_y = data[12]
        self.pos_z = data[13]
        self.location = (data[11], data[12], data[13])  # location (x,y,z) in camera coord.
        self.ry = data[14]  # yaw angle (around Y-axis in camera coordinates) [-pi..pi]

    def estimate_diffculty(self):
        """ Function that estimate difficulty to detect the object as defined in kitti website"""
        # height of the bounding box
        bb_height = np.abs(self.xmax - self.xmin)

        if bb_height >= 40 and self.occlusion == 0 and self.truncation <= 0.15:
            return "Easy"
        elif bb_height >= 25 and self.occlusion in [0, 1] and self.truncation <= 0.30:
            return "Moderate"
        elif ( bb_height >= 25 and self.occlusion in [0, 1, 2] and self.truncation <= 0.50):
            return "Hard"
        else:
            return "Unknown"

    def print_object(self):
        print(f"Type, truncation, occlusion, alpha: {self.type}, {self.truncation}, {self.occlusion}, {self.alpha}")
        print(f"2d bbox (x0,y0,x1,y1): {self.xmin}, {self.ymin}, {self.xmax}, {self.ymax}")
        print(f"3d bbox h,w,l: {self.height}, {self.width}, {self.length}")
        print(f"3d bbox location, ry: ({self.pos_x}, {self.pos_y}, {self.pos_z}), {self.ry}")
        print("Difficulty of estimation:{}".format(self.estimate_diffculty()))

class Calibration(object):
    """ Calibration matrices and utils
        3d XYZ in <label>.txt are in rect camera coord.
        2d box xy are in image2 coord
        Points in <lidar>.bin are in Velodyne coord.

        y_image2 = P^2_rect * x_rect
        y_image2 = P^2_rect * R0_rect * Tr_velo_to_cam * x_velo
        x_ref = Tr_velo_to_cam * x_velo
        x_rect = R0_rect * x_ref

        P^2_rect = [f^2_u,  0,      c^2_u,  -f^2_u b^2_x;
                    0,      f^2_v,  c^2_v,  -f^2_v b^2_y;
                    0,      0,      1,      0]
                 = K * [1|t]

        image2 coord:
         ----> x-axis (u)
        |
        |
        v y-axis (v)

        velodyne coord:
        front x, left y, up z

        rect/ref camera coord:
        right x, down y, front z

        Ref (KITTI paper): http://www.cvlibs.net/publications/Geiger2013IJRR.pdf

        TODO(rqi): do matrix multiplication only once for each projection.
    """

    def __init__(self, calib_filepath, from_video=False):
        if from_video:
            calibs = self.read_calib_from_video(calib_filepath)
        else:
            calibs = self.read_calib_file(calib_filepath)
        # Projection matrix from rect camera coord to image2 coord
        self.P = calibs["P2"]
        self.P = np.reshape(self.P, [3, 4])
        # Rigid transforms from Velodyne coord to reference camera coord
        self.V2C = calibs["Tr_velo_to_cam"]
        self.V2C = np.reshape(self.V2C, [3, 4])
        self.C2V = inverse_rigid_trans(self.V2C)
        # Rotation from reference camera coord to rect camera coord
        self.R0 = calibs["R0_rect"]
        self.R0 = np.reshape(self.R0, [3, 3])

        # Camera intrinsics and extrinsics
        self.c_u = self.P[0, 2]
        self.c_v = self.P[1, 2]
        self.f_u = self.P[0, 0]
        self.f_v = self.P[1, 1]
        self.b_x = self.P[0, 3] / (-self.f_u)  # relative
        self.b_y = self.P[1, 3] / (-self.f_v)

    def read_calib_file(self, filepath):
        """ Read in a calibration file and parse into a dictionary.
        Ref: https://github.com/utiasSTARS/pykitti/blob/master/pykitti/utils.py
        """
        data = {}
        with open(filepath, "r") as f:
            for line in f.readlines():
                line = line.rstrip()
                if len(line) == 0:
                    continue
                key, value = line.split(":", 1)
                # The only non-float values in these files are dates, which
                # we don't care about anyway
                try:
                    data[key] = np.array([float(x) for x in value.split()])
                except ValueError:
                    pass

        return data

    def read_calib_from_video(self, calib_root_dir):
        """ Read calibration for camera 2 from video calib files.
            there are calib_cam_to_cam and calib_velo_to_cam under the calib_root_dir
        """
        data = {}
        cam2cam = self.read_calib_file(
            os.path.join(calib_root_dir, "calib_cam_to_cam.txt")
        )
        velo2cam = self.read_calib_file(
            os.path.join(calib_root_dir, "calib_velo_to_cam.txt")
        )
        Tr_velo_to_cam = np.zeros((3, 4))
        Tr_velo_to_cam[0:3, 0:3] = np.reshape(velo2cam["R"], [3, 3])
        Tr_velo_to_cam[:, 3] = velo2cam["T"]
        data["Tr_velo_to_cam"] = np.reshape(Tr_velo_to_cam, [12])
        data["R0_rect"] = cam2cam["R_rect_00"]
        data["P2"] = cam2cam["P_rect_02"]
        return data

    def cart2hom(self, pts_3d):
        """ Input: nx3 points in Cartesian
            Oupput: nx4 points in Homogeneous by pending 1
        """
        n = pts_3d.shape[0]
        pts_3d_hom = np.hstack((pts_3d, np.ones((n, 1))))
        return pts_3d_hom

    # ===========================
    # ------- 3d to 3d ----------
    # ===========================
    def project_velo_to_ref(self, pts_3d_velo):
        pts_3d_velo = self.cart2hom(pts_3d_velo)  # nx4
        return np.dot(pts_3d_velo, np.transpose(self.V2C))

    def project_ref_to_velo(self, pts_3d_ref):
        pts_3d_ref = self.cart2hom(pts_3d_ref)  # nx4
        return np.dot(pts_3d_ref, np.transpose(self.C2V))

    def project_rect_to_ref(self, pts_3d_rect):
        """ Input and Output are nx3 points """
        return np.transpose(np.dot(np.linalg.inv(self.R0), np.transpose(pts_3d_rect)))

    def project_ref_to_rect(self, pts_3d_ref):
        """ Input and Output are nx3 points """
        return np.transpose(np.dot(self.R0, np.transpose(pts_3d_ref)))

    def project_rect_to_velo(self, pts_3d_rect):
        """ Input: nx3 points in rect camera coord.
            Output: nx3 points in velodyne coord.
        """
        pts_3d_ref = self.project_rect_to_ref(pts_3d_rect)
        return self.project_ref_to_velo(pts_3d_ref)

    def project_velo_to_rect(self, pts_3d_velo):
        pts_3d_ref = self.project_velo_to_ref(pts_3d_velo)
        return self.project_ref_to_rect(pts_3d_ref)

    # ===========================
    # ------- 3d to 2d ----------
    # ===========================
    def project_rect_to_image(self, pts_3d_rect):
        """ Input: nx3 points in rect camera coord.
            Output: nx2 points in image2 coord.
        """
        pts_3d_rect = self.cart2hom(pts_3d_rect)
        pts_2d = np.dot(pts_3d_rect, np.transpose(self.P))  # nx3
        pts_2d[:, 0] /= pts_2d[:, 2]
        pts_2d[:, 1] /= pts_2d[:, 2]
        return pts_2d[:, 0:2]

    def project_velo_to_image(self, pts_3d_velo):
        """ Input: nx3 points in velodyne coord.
            Output: nx2 points in image2 coord.
        """
        pts_3d_rect = self.project_velo_to_rect(pts_3d_velo)
        return self.project_rect_to_image(pts_3d_rect)

    def project_8p_to_4p(self, pts_2d):
        x0 = np.min(pts_2d[:, 0])
        x1 = np.max(pts_2d[:, 0])
        y0 = np.min(pts_2d[:, 1])
        y1 = np.max(pts_2d[:, 1])
        x0 = max(0, x0)
        # x1 = min(x1, proj.image_width)
        y0 = max(0, y0)
        # y1 = min(y1, proj.image_height)
        return np.array([x0, y0, x1, y1])

    def project_velo_to_4p(self, pts_3d_velo):
        """ Input: nx3 points in velodyne coord.
            Output: 4 points in image2 coord.
        """
        pts_2d_velo = self.project_velo_to_image(pts_3d_velo)
        return self.project_8p_to_4p(pts_2d_velo)

    # ===========================
    # ------- 2d to 3d ----------
    # ===========================
    def project_image_to_rect(self, uv_depth):
        """ Input: nx3 first two channels are uv, 3rd channel
                   is depth in rect camera coord.
            Output: nx3 points in rect camera coord.
        """
        n = uv_depth.shape[0]
        x = ((uv_depth[:, 0] - self.c_u) * uv_depth[:, 2]) / self.f_u + self.b_x
        y = ((uv_depth[:, 1] - self.c_v) * uv_depth[:, 2]) / self.f_v + self.b_y
        pts_3d_rect = np.zeros((n, 3))
        pts_3d_rect[:, 0] = x
        pts_3d_rect[:, 1] = y
        pts_3d_rect[:, 2] = uv_depth[:, 2]
        return pts_3d_rect

    def project_image_to_velo(self, uv_depth):
        pts_3d_rect = self.project_image_to_rect(uv_depth)
        return self.project_rect_to_velo(pts_3d_rect)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/2 17:28:03

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import matplotlib.pyplot as plt
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = [
    'draw_bar',
    'draw_bar_h',
    'draw_barbs',
    'draw_boxplot',
    'draw_contour',
    'draw_errorbar',
    'draw_event',
    'draw_fill',
    'draw_fill_between',
    'draw_fill_between_x',
    'draw_hexbin',
    'draw_hist',
    'draw_image',
    'draw_line',
    'draw_pie',
    'draw_polar',
    'draw_quiver',
    'draw_scatter',
    'draw_stack',
    'draw_stem',
    'draw_step',
    'draw_text',
    'draw_violin',
    'draw_xcoor',
]

draw_bar = plt.bar              # 垂直柱形图
draw_bar_h = plt.barh           # 水平条形图

draw_barbs = plt.barbs

draw_boxplot = plt.boxplot  # 箱型图

draw_contour = plt.contour

draw_errorbar = plt.errorbar

draw_event = plt.eventplot

draw_fill = plt.fill

draw_fill_between = plt.fill_between

draw_fill_between_x = plt.fill_betweenx

draw_hexbin = plt.hexbin

draw_hist = plt.hist            # 直方图

draw_hist2d = plt.hist2d        # 2D直方图

draw_image = plt.imshow         # 画图

draw_line = plt.plot            # 线条

draw_pie = plt.pie              # 饼图

draw_polar = plt.polar        # 极坐标图

draw_quiver = plt.quiver    # 二维箭头图

draw_scatter = plt.scatter      # 绘制x与y的散点图

draw_stack = plt.stackplot  # 绘制堆叠图

draw_stem = plt.stem    # 针状图（又称为“火柴图”）

draw_step = plt.step    # 阶梯图

draw_text = plt.text    # 文字

draw_violin = plt.violinplot

draw_xcoor = plt.xcorr


# import laok; laok.dump_module_all(preffix='draw_')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/3 21:26:54

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import laok
from sklearn import datasets
# ===============================================================================
r'''数据集定义
'''
# ===============================================================================
__all__ = [
 'data_make_biclusters',
 'data_make_blobs',
 'data_make_checkerboard',
 'data_make_circles',
 'data_make_classification',
 'data_make_friedman1',
 'data_make_friedman2',
 'data_make_friedman3',
 'data_make_gaussian_quantiles',
 'data_make_hastie_10_2',
 'data_make_low_rank_matrix',
 'data_make_moons',
 'data_make_multilabel_classification',
 'data_make_regression',
 'data_make_s_curve',
 'data_make_sparse_coded_signal',
 'data_make_sparse_spd_matrix',
 'data_make_sparse_uncorrelated',
 'data_make_spd_matrix',
 'data_make_swiss_roll',


 'data_breast_cancer',
 'data_diabetes',
 'data_digits',
 'data_files',
 'data_iris',
 'data_linnerud',
 'data_sample_image',
 'data_sample_images',
 'data_svmlight_file',
 'data_svmlight_files',
 'data_wine',
]



data_make_biclusters = datasets.make_biclusters
data_make_blobs = datasets.make_blobs
data_make_checkerboard = datasets.make_checkerboard
data_make_circles = datasets.make_circles
data_make_classification = datasets.make_classification
data_make_friedman1 = datasets.make_friedman1
data_make_friedman2 = datasets.make_friedman2
data_make_friedman3 = datasets.make_friedman3
data_make_gaussian_quantiles = datasets.make_gaussian_quantiles
data_make_hastie_10_2 = datasets.make_hastie_10_2
data_make_low_rank_matrix = datasets.make_low_rank_matrix
data_make_moons = datasets.make_moons
data_make_multilabel_classification = datasets.make_multilabel_classification
data_make_regression = datasets.make_regression
data_make_s_curve = datasets.make_s_curve
data_make_sparse_coded_signal = datasets.make_sparse_coded_signal
data_make_sparse_spd_matrix = datasets.make_sparse_spd_matrix
data_make_sparse_uncorrelated = datasets.make_sparse_uncorrelated
data_make_spd_matrix = datasets.make_spd_matrix
data_make_swiss_roll = datasets.make_swiss_roll

data_breast_cancer = datasets.load_breast_cancer
data_diabetes = datasets.load_diabetes
data_digits = datasets.load_digits
data_files = datasets.load_files
data_iris = datasets.load_iris
data_linnerud = datasets.load_linnerud
data_sample_image = datasets.load_sample_image
data_sample_images = datasets.load_sample_images
data_svmlight_file = datasets.load_svmlight_file
data_svmlight_files = datasets.load_svmlight_files
data_wine = datasets.load_wine





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/9 19:43:59

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['apply_batch', 'apply_sample_point', 'apply_sample_points', 'mat_angle_axis', 'plane_symm']

def apply_batch(batch_data, trans):
    B = batch_data.shape[0]
    for k in range(B):
        batch_data[k, ...] = trans(batch_data[k, ...])
    return batch_data

def apply_sample_points(points, trans, ratio=0.8):
    sample_ratio = np.random.random() * ratio
    drop_idx = np.where(np.random.random((points.shape[0])) <= sample_ratio)[0]
    if len(drop_idx) > 0:
        points[drop_idx] = trans(points[drop_idx])
    return points

def apply_sample_point(points, trans, ratio=0.8):
    sample_ratio = np.random.random() * ratio
    drop_idx = np.where(np.random.random((points.shape[0])) <= sample_ratio)[0]
    for idx in drop_idx:
        points[idx] = trans(points[idx])
    return points

def mat_angle_axis(angle: float, axis: np.ndarray):
    u = axis / np.linalg.norm(axis)
    cosval, sinval = np.cos(angle), np.sin(angle)
    # yapf: disable
    cross_prod_mat = np.array([[0.0, -u[2], u[1]],
                               [u[2], 0.0, -u[0]],
                               [-u[1], u[0], 0.0]])

    R = cosval * np.eye(3) + sinval * cross_prod_mat + (1.0 - cosval) * np.outer(u, u)
    # yapf: enable
    return R

def plane_symm(points, a=0, b=0, c=0, d=0):
    if a == 0 and b == 0 and c == 0:
        a = 0.0001
    if len(points.shape) > 1:
        lam = ((a * points[:, 0] + b * points[:, 1] + c * points[:, 2] + d) / (a * a + b * b + c * c))
        points[:, 0] -= 2 * lam * a
        points[:, 1] -= 2 * lam * b
        points[:, 2] -= 2 * lam * c
    else:
        lam = ((a * points[0] + b * points[1] + c * points[2] + d) / (a * a + b * b + c * c))
        points[0] -= 2 * lam * a
        points[1] -= 2 * lam * b
        points[2] -= 2 * lam * c
    return points



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/19 22:34:15
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from .log import get_logger
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['log_debug', 'log_info', 'log_warning', 'log_error', 'log_critical']

#默认日志
_logger = get_logger()

# 日志函数
log_debug = _logger.debug
log_info = _logger.info
log_warning = _logger.warning
log_error = _logger.error
log_critical = _logger.critical



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/16 01:18:03

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['load_cld_xyz', 'load_cld_bin', 'save_cld_xyz']

def load_cld_xyz(filename, delimiter=',', **kws):
    return np.loadtxt(filename, delimiter=delimiter, **kws)

def save_cld_xyz(filename, points, delimiter=',', **kws):
    np.savetxt(filename, points, delimiter=delimiter, **kws)

def load_cld_bin(filename, feature=4):
    return np.fromfile(filename, dtype=np.float32).reshape((-1, feature))




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/19 22:43:12
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from flask import Flask, request
#===============================================================================
r'''web服务
'''
#===============================================================================
__all__ = ['app', 'run', 'req_param', 'req_files']

app = Flask(__name__)

run = app.run

def req_param(name=None):
    if request.method == 'POST':
        data = request.form
    else:
        data = request.args

    if name is None:
        return data
    return data.get(name)

def req_files():
    return



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/12/21 00:28:06

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from lxml import etree
import json, base64
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['voc_to_labelme']

def voc_to_labelme(xml_file, needImageData = False):
    with open(xml_file, mode='rb') as f:
        xml_data = etree.XML(f.read())
    jdata = {}
    jdata['version'] = "5.0.1"
    jdata["flags"] = {}
    shapes = []
    for obj in xml_data.findall('object'):
        name = obj.find('name').text
        xmin = int(obj.find('bndbox/xmin').text)
        ymin = int(obj.find('bndbox/ymin').text)
        xmax = int(obj.find('bndbox/xmax').text)
        ymax = int(obj.find('bndbox/ymax').text)
        shapes.append({
            "label": name,
            "points":[
                [ xmin, ymin ],
                [ xmax, ymax ]
            ],
            "group_id": {},
            "shape_type": "rectangle",
            "flags": {}
        })
    jdata['shapes'] = shapes

    imagePath = xml_data.find('filename').text
    jdata['imagePath'] = imagePath
    width = xml_data.find('size/width').text
    height = xml_data.find('size/height').text

    if needImageData:
        with open(imagePath, 'rb') as f:
            imageData = f.read()
        jdata['imageData'] = base64.b64encode(imageData).decode('utf-8')
    else:
        jdata['imageData'] = None
    jdata['imageHeight'] = int(height)
    jdata['imageWidth'] = int(width)

    json_file = xml_file.replace('.xml', '.json')
    with open(json_file, 'w', encoding='utf8') as wf:
        json.dump(jdata, wf, indent=2)
    return jdata



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/4/19 13:51:55

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import threading
import concurrent.futures as futures
from ..fs import (files_under, files_cur, dirs_cur, dirs_under)
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['thread_id', 'thread_count', 'paral_seq', 'paral_files_under', 'paral_files_cur', 'paral_dirs_under', 'paral_dirs_cur', ]

def thread_id():
    return threading.get_ident()

def thread_count():
    return threading.active_count()

####################    并行接口
def paral_seq(data_list, handler, max_workers=None, use_thread = True):
    Executor = futures.ThreadPoolExecutor if use_thread else futures.ProcessPoolExecutor
    with Executor(max_workers=max_workers) as ex:
        return [val for val in ex.map(handler, data_list)]

def paral_files_under(dir_name, handler, suffix_list = None, max_workers=None, use_thread = True):
    data_iter = files_under(dir_name, suffix_list)
    return paral_seq(data_iter, handler, max_workers=max_workers, use_thread=use_thread)

def paral_files_cur(dir_name, handler, suffix_list = None, max_workers=None, use_thread = True):
    data_iter = files_cur(dir_name, suffix_list)
    return paral_seq(data_iter, handler, max_workers=max_workers, use_thread=use_thread)

def paral_dirs_under(dir_name, handler, suffix_list = None, max_workers=None, use_thread = True):
    data_iter = dirs_under(dir_name, suffix_list)
    return paral_seq(data_iter, handler, max_workers=max_workers, use_thread=use_thread)

def paral_dirs_cur(dir_name, handler, suffix_list = None, max_workers=None, use_thread = True):
    data_iter = dirs_cur(dir_name, suffix_list)
    return paral_seq(data_iter, handler, max_workers=max_workers, use_thread=use_thread)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/9 19:27:37

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
import matplotlib.pyplot as plt
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['plot_acc', ]

def plot_acc(train_acc, test_acc, title='train and test accuracy', save_img_file=None, plt_show=True):
    epochs = list(range(1, len(train_acc) + 1))

    # 创建图形和坐标轴
    plt.figure(figsize=(10, 6))

    # 绘制训练准确率曲线
    train_line, = plt.plot(epochs, train_acc, label='train')
    max_train_idx = train_acc.index(max(train_acc))
    plt.annotate(f'max: {max(train_acc):.4f}',
                 xy=(epochs[max_train_idx], train_acc[max_train_idx]),
                 xytext=(epochs[max_train_idx]+0.1, train_acc[max_train_idx]+0.01),
                 # arrowprops=dict(facecolor='black', shrink=0.05, width=1.5, headwidth=8)
                 )
    plt.scatter([epochs[max_train_idx]], [train_acc[max_train_idx]], color='red', s=50)

    # 绘制测试准确率曲线
    test_line, = plt.plot(epochs, test_acc,  label='test')
    max_test_idx = test_acc.index(max(test_acc))
    plt.annotate(f'max: {max(test_acc):.4f}',
                 xy=(epochs[max_test_idx], test_acc[max_test_idx]),
                 xytext=(epochs[max_test_idx]+0.1, test_acc[max_test_idx]+0.01),
                 # arrowprops=dict(facecolor='black', shrink=0.05, width=1.5, headwidth=8)
                 )
    plt.scatter([epochs[max_test_idx]], [test_acc[max_test_idx]], color='red', s=50)

    # 设置图表标题和坐标轴标签
    plt.title(title)
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    # plt.xticks(epochs)  # 设置x轴刻度为整数轮次
    plt.grid(True, linestyle='--', alpha=0.7)

    # 添加图例和显示图形
    plt.legend(handles=[train_line, test_line])
    plt.tight_layout()  # 确保标签和注释不会超出图形边界

    if save_img_file:
        plt.savefig(save_img_file)

    if plt_show:
        plt.show()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/6/4 18:13:09

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import inspect, contextlib, types, sys
#===============================================================================
# 
#===============================================================================

__all__ = ['dump', 'dump_help', 'is_method_overide', 'get_members', 'get_module_path']

_MODULE_DEFAULT_FIELDS = ['__builtins__', '__cached__', '__doc__', '__file__',
                          '__loader__', '__name__', '__package__', '__path__', '__spec__']

_CLASS_DEFAULT_FIELDS = ['__dict__', '__doc__', '__module__', '__weakref__',
                         '__init_subclass__', '__subclasshook__', '__class__']

def get_module_path(mod):
    pth = ''

    fname = getattr(mod, '__file__', '')
    if fname:
        pth = fname

    nsPath = getattr(mod, '__path__', None)
    if nsPath:
        pth = nsPath[0]
    return pth

def _is_third_pkg(mod):
    return 'site-packages' in get_module_path(mod)

def _is_in_same_pkg(mod1, mod2):
    mod1 = get_module_path(mod1)
    mod2 = get_module_path(mod2)
    return mod1 in mod2

def is_method_overide(obj, method_name, parent=None):
    '''判断方法是否覆盖父类的方法'''
    if parent is None:
        parent = object

    mth = getattr(obj, method_name)
    if hasattr(parent, method_name):
        p_mth_id = id(getattr(parent, method_name))
        if isinstance(obj, type):  # 如果是类
            if id(mth) == p_mth_id:
                return True
        else:  # 如果是对象
            if id(getattr(obj.__class__, method_name)) == p_mth_id:
                return True
    return False

def dump_help(obj):
    with open('help-%s.txt' % _val_name(obj), 'w') as f, \
            contextlib.redirect_stdout(f):
        help(obj)

def dump(obj, ignore_fields = tuple(),
                use_all = True,
                dump_format = None,
                skip_moudle_defaults = True,
                skip_class_defaults=True,
                skip_underscore=True,
                stream=None):

    # # 空对象,直接不处理
    # if cfg is None:
    #     print('None')
    #     return

    all_ModuleType = [] #模块
    all_Class = []      #类
    all_Attr = []       #属性
    all_Funcs = []      #函数
    if stream is None:
        stream = sys.stdout
    name_value_iter = get_members(obj, ignore_fields=ignore_fields,
                                  use_all=use_all, skip_moudle_defaults=skip_moudle_defaults,
                                  skip_class_defaults=skip_class_defaults,
                                  skip_underscore=skip_underscore)
    stream.write(f'\n########## {type(obj)} ##########\n')
    if dump_format is not None:
        stream.write(f'from {obj.__name__} import *\n')

    for item in name_value_iter:
        k, v = item
        if isinstance(v, type): # class类型
            all_Class.append(item)

        elif isinstance(v, types.ModuleType): # 模块类型
            if not _is_third_pkg(v): # 保证是第三方库
                continue
            if isinstance(obj, types.ModuleType): #保证在同一个包里面
                if not _is_in_same_pkg(obj, v):
                    continue
            all_ModuleType.append(item)

        elif isinstance(v, (types.FunctionType,types.MethodType,types.WrapperDescriptorType,
                            types.MethodWrapperType, types.MethodDescriptorType, types.BuiltinFunctionType,
                            types.BuiltinMethodType) #可执行的函数类型
                        ):
            all_Funcs.append(item)
        else:
            all_Attr.append(item)

    if all_ModuleType:
        stream.write('##### module\n')
        for k,v in all_ModuleType:
            if dump_format is None:
                stream.write('  # %s[%s]\n' % (k, _val_doc(v)))
            else:
                stream.write('  %s #[%s]\n' % (k, _val_doc(v)))

    if all_Class:
        stream.write('##### class\n')
        for k,v in all_Class:
            if dump_format is None:
                stream.write('  # %s[%s]\n' % (k, _val_doc(v)))
            else:
                stream.write('  %s #[%s]\n' % (k, _val_doc(v)))

    if all_Attr:
        stream.write('##### attr\n')
        for k,v in all_Attr:
            if dump_format is None or (k.startswith('__') and k.endswith('__')):
                stream.write('  # %s[%s] [%s]\n' % (k, _val_str(v), _val_doc(v)))
            else:
                stream.write('  %s #[%s] [%s]\n' % (k, _val_str(v), _val_doc(v)))

    if all_Funcs:
        stream.write('##### func\n')
        for k,v in all_Funcs:
            if dump_format is None:
                stream.write('  # %s%s [%s]\n' % (k, _func_signature(v), _val_doc(v)))
            else:
                stream.write('  %s #%s [%s]\n' % (k, _func_signature(v), _val_doc(v)))
    stream.write("\n")

def _val_str(v):
    try:
        return str(v).replace('\n', ' ').replace('\r', ' ')[0:100]
    except Exception as e:
        return ""
def _val_doc(v):
    if isinstance(v, str):
        return  '`str`'
    if isinstance(v, bytes):
        return  '`bytes`'
    elif isinstance(v, int):
        return  '`integer`'
    elif isinstance(v, dict):
        return  '`dict`'
    elif isinstance(v, set):
        return  '`set`'
    elif isinstance(v, frozenset):
        return '`fronzenset`'
    elif isinstance(v, float):
        return '`float`'
    elif isinstance(v, tuple):
        return '`tuple`'
    elif isinstance(v, list):
        return '`list`'

    try:
        import re
        doc = inspect.getdoc(v)
        doc = re.sub(' +', ' ', doc)
        return doc.replace('\n', ' ').replace('\r', ' ')[0:250]
    except Exception as e:
        return  ''

def _val_name(v):
    try:
        name = v.__name__
    except Exception:
        try:
            name = v.__class__.__name__
        except Exception:
            name = 'none'
    return name

def _func_signature(v):
    try:
        return str(inspect.signature(v))
    except: # 有些函数没有签名
        return "(...)"

def get_members(obj,
                ignore_fields = tuple(),
                use_all = True,
                skip_moudle_defaults = True,
                skip_class_defaults=True,
                skip_underscore=True):
    '''
    获取可能的属性列表
    '''

    # 默认从 __all__获取属性
    datalist = []
    if use_all:
        for k in getattr(obj, '__all__', tuple()):
            if hasattr(obj, k):
                datalist.append( (k, getattr(obj, k)) )

    # 获取所有属性
    if not datalist: # 如果 __all__是空的,就从模块里获取
        datalist = _getmembers(obj)

    if isinstance(ignore_fields, str):
        ignore_fields = ignore_fields.split()

    datalist.sort(key=lambda kv: kv[0])

    # 把 __init__放在最前面
    for i,(k,v) in enumerate(datalist):
        if k == '__init__':
            del datalist[i]
            datalist.insert(0, (k, v))
            break

    #字段过滤
    for k,v in datalist:

        # 忽略的字段
        if k in ignore_fields:
            continue

        # 私有变量不访问
        if skip_underscore and k.startswith('_') :
            if not k.startswith('__') or not k.endswith('__'):
                continue

        # 跳过模块中方法
        if skip_moudle_defaults and isinstance(obj, types.ModuleType):
            if k in _MODULE_DEFAULT_FIELDS:
                continue

        if skip_class_defaults and hasattr(obj, '__class__'):

            # 跳过默认的参数
            if k in _CLASS_DEFAULT_FIELDS:
                continue

            # 跳过 object 中没有覆盖的方法
            if is_method_overide(obj, k):
                continue

        yield k,v

    return datalist

def _getmembers(object, predicate=None):

    if inspect.isclass(object):
        mro = (object,) + inspect.getmro(object)
    else:
        mro = ()

    results = []
    processed = set()
    names = dir(object)

    try:
        for base in object.__bases__:
            for k, v in base.__dict__.items():
                if isinstance(v, types.DynamicClassAttribute):
                    names.append(k)
    except AttributeError:
        pass

    for key in names:
        try:
            value = getattr(object, key)
            if key in processed:
                raise AttributeError
        except AttributeError:
            for base in mro:
                if key in base.__dict__:
                    value = base.__dict__[key]
                    break
            else:
                continue
        except Exception:
            continue

        if not predicate or predicate(value):
            results.append((key, value))
        processed.add(key)

    return results



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/19 23:05:22
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from ..app import app
from flask import jsonify, request, g, session, current_app
#===============================================================================
r'''
'''
#===============================================================================

@app.route("/", methods=['POST', 'GET'])
def index():
    return jsonify(data = "Wellcom to laok's World!", status='success')

@app.errorhandler(404)
def not_found(error):
    return jsonify(data = 'not found', status='fail'), 404

@app.route('/debug', methods=['post', 'get', 'head', 'options', 'put', 'patch', 'delete'])
def requests():
    return jsonify(requests)



# EMD approximation module (based on auction algorithm)
# memory complexity: O(n)
# time complexity: O(n^2 * iter) 
# author: Minghua Liu

# Input:
# xyz1, xyz2: [#batch, #points, 3]
# where xyz1 is the predicted point cloud and xyz2 is the ground truth point cloud 
# two point clouds should have same size and be normalized to [0, 1]
# #points should be a multiple of 1024
# #batch should be no greater than 512
# eps is a parameter which balances the error rate and the speed of convergence
# iters is the number of iteration
# we only calculate gradient for xyz1

# Output:
# dist: [#batch, #points],  sqrt(dist) -> L2 distance 
# assignment: [#batch, #points], index of the matched point in the ground truth point cloud
# the result is an approximation and the assignment is not guranteed to be a bijection
import torch
from torch import nn
from torch.autograd import Function
import emd


class emdFunction(Function):
    @staticmethod
    def forward(ctx, xyz1, xyz2, eps, iters):

        batchsize, n, _ = xyz1.size()
        _, m, _ = xyz2.size()

        assert(n == m)
        assert(xyz1.size()[0] == xyz2.size()[0])
        assert(n % 1024 == 0)
        assert(batchsize <= 512)

        xyz1 = xyz1.contiguous().float().cuda()
        xyz2 = xyz2.contiguous().float().cuda()
        dist = torch.zeros(batchsize, n, device='cuda').contiguous()
        assignment = torch.zeros(batchsize, n, device='cuda', dtype=torch.int32).contiguous() - 1
        assignment_inv = torch.zeros(batchsize, m, device='cuda', dtype=torch.int32).contiguous() - 1
        price = torch.zeros(batchsize, m, device='cuda').contiguous()
        bid = torch.zeros(batchsize, n, device='cuda', dtype=torch.int32).contiguous()
        bid_increments = torch.zeros(batchsize, n, device='cuda').contiguous()
        max_increments = torch.zeros(batchsize, m, device='cuda').contiguous()
        unass_idx = torch.zeros(batchsize * n, device='cuda', dtype=torch.int32).contiguous()
        max_idx = torch.zeros(batchsize * m, device='cuda', dtype=torch.int32).contiguous()
        unass_cnt = torch.zeros(512, dtype=torch.int32, device='cuda').contiguous()
        unass_cnt_sum = torch.zeros(512, dtype=torch.int32, device='cuda').contiguous()
        cnt_tmp = torch.zeros(512, dtype=torch.int32, device='cuda').contiguous()

        emd.forward(xyz1, xyz2, dist, assignment, price, assignment_inv, bid, bid_increments, max_increments, unass_idx, unass_cnt, unass_cnt_sum, cnt_tmp, max_idx, eps, iters)

        ctx.save_for_backward(xyz1, xyz2, assignment)
        return dist, assignment

    @staticmethod
    def backward(ctx, graddist, gradidx):
        xyz1, xyz2, assignment = ctx.saved_tensors
        graddist = graddist.contiguous()

        gradxyz1 = torch.zeros(xyz1.size(), device='cuda').contiguous()
        gradxyz2 = torch.zeros(xyz2.size(), device='cuda').contiguous()

        emd.backward(xyz1, xyz2, gradxyz1, graddist, assignment)
        return gradxyz1, gradxyz2, None, None

class emdModule(nn.Module):
    def __init__(self):
        super(emdModule, self).__init__()

    def forward(self, input1, input2, eps, iters):
        return emdFunction.apply(input1, input2, eps, iters)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/7/12 22:12:50

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['edge_sobel_x', 'edge_sobel_y', 'edge_sobel',
            'edge_scharr_x', 'edge_scharr_y', 'edge_scharr',
           'edge_laplace', 'edge_canny',
]

def edge_sobel_x(img):
    pass

def edge_sobel_y(img):
    pass

def edge_sobel(img):
    pass

def edge_scharr_x(img):
    pass

def edge_scharr_y(img):
    pass

def edge_scharr(img):
    pass

def edge_laplace(img):
    pass

def edge_canny(img, minVal=50, maxVal=150, apertureSize=3, L2gradient=False):
    return cv2.Canny(img, threshold1=minVal, threshold2=maxVal, apertureSize=apertureSize, L2gradient=L2gradient)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/16 14:54:01

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import torch
import laok
from laok.ext.torch_.io import load_model_params
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['CheckPointBest', 'select_best_model_file', 'load_best_model']

####################    支持搜索加载最优的参数模型
class CheckPointBest:
    def __init__(self, model, save_dir, tag=None, load_path=True, keep_best=False, show_info=True):
        self._save_dir = save_dir if save_dir else "out/" + laok.datetime_fmt()
        if save_dir is None:
            self._save_dir = "out/" + laok.datetime_fmt()
            self._tag = 'model' if tag is None else tag
        else:
            self._save_dir = save_dir
            self._tag = laok.path_stem(save_dir) if tag is None else tag

        self._keep_best = keep_best
        self.show_info = show_info
        if self.show_info:
            laok.log_info(f'===== {self.__class__.__name__}')
            laok.log_info(f"tag: {self._tag}, keep_best:{self._keep_best}, path:{self._save_dir}")

        self._name_value = {}
        self._value = None
        self._is_keep_best = False
        self.set_model(model)
        if load_path:
            self.load_from_path()

    def set_model(self, model):
        self._model = model

    def update(self, value):
        '''the model needs to be saved according to the value
        '''
        # check if need save model
        if self._value is None or self._value < value:
            self._value = value  # update the best value
            save_file = laok.path_join(self._save_dir, f'{self._tag}-{self._value:.4f}.pth', make_parent=True)
            torch.save(self._model.state_dict(), save_file)
            if self.show_info:
                laok.log_info(f'save checkpoint: {save_file}')

        if self._value == value:
            if self.show_info:
                laok.log_info(f"current {self._tag} value is {value:.4f}, it's the best.!!!")
        else:
            if self.show_info:
                laok.log_info(f'current {self._tag} value is {value:.4f}, while the best value is {self._value:.4f}')
            if self._keep_best:
                self._is_keep_best = True
                if self.show_info:
                    laok.log_info(f"keep best")
                self.load_from_path()
                self._is_keep_best = False

    def load_from_path(self):
        max_value_file, max_value = _select_best_mode_file(self._save_dir)
        if max_value_file:
            self.load_from_file(max_value_file, max_value)

    def load_from_file(self, model_file, value=None):
        if not os.path.exists(model_file):
            return
        if self.show_info:
            laok.log_info(f'load checkpoint file: {model_file}')
        load_model_params(self._model, model_file)
        _value = _score_from_path(model_file)
        self._value = _value if value is None else value

def select_best_model_file(dir):
    filename, val = _select_best_mode_file(dir)
    return filename

def load_best_model(model, dir, show_info=True):
    model_file = select_best_model_file(dir)
    if show_info:
        laok.log_info(f'load checkpoint file: {model_file}')
    load_model_params(model, model_file)
    return model

def _select_best_mode_file(dir):
    max_value = None
    max_value_file = None

    for fname in laok.files_under(dir, '.pth'):
        _value = _score_from_path(fname)
        if max_value is None or max_value < _value:
            max_value = _value
            max_value_file = fname

    return max_value_file, max_value

def _score_from_path(filename):
    stem = laok.path_stem(filename)
    value = stem.split('-')[-1]
    return laok.float_value(value, 0.0)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/9/17 23:37:17

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import numpy as np
from torch.utils.data import Dataset
import laok.util.log as klog
import laok.util.fs as kfs
import laok.util.parallel as kparal
import laok.cv3d as kcv3
# ===============================================================================
r'''
'''
# ===============================================================================

class ModelnetNormalsDiff(Dataset):
    def __init__(self, root, split='train',
                 num_point = 1024,
                 num_category = 40,
                 normal_small_knn = 3,
                 normal_large_knn = 6,
                 transform=None,
                 cached = False,
                 show_info = False,
                 ):
        self.root = root
        self.npoints = num_point
        self.num_category = num_category
        self.transform = transform
        self.split = split
        klog.debug(f'===== {self.__class__.__name__}')
        klog.debug(f'root={root}')
        klog.debug(f'split={split}, num_point={num_point}, num_category={num_category}')

        if self.num_category == 10:
            self.catfile = os.path.join(self.root, 'modelnet10_shape_names.txt')
        else:
            self.catfile = os.path.join(self.root, 'modelnet40_shape_names.txt')

        klog.debug(f'catfile={self.catfile}')

        self.cat = [line.rstrip() for line in open(self.catfile)]
        self.classes = dict(zip(self.cat, range(len(self.cat))))

        klog.debug(f'cat={self.cat}')
        klog.debug(f'classes={self.classes}')

        shape_ids = {}
        if self.num_category == 10:
            shape_ids['train'] = [line.rstrip() for line in open(os.path.join(self.root, 'modelnet10_train.txt'))]
            shape_ids['test'] = [line.rstrip() for line in open(os.path.join(self.root, 'modelnet10_test.txt'))]
        else:
            shape_ids['train'] = [line.rstrip() for line in open(os.path.join(self.root, 'modelnet40_train.txt'))]
            shape_ids['test'] = [line.rstrip() for line in open(os.path.join(self.root, 'modelnet40_test.txt'))]

        assert (split == 'train' or split == 'test')
        shape_names = ['_'.join(x.split('_')[0:-1]) for x in shape_ids[split]]
        self.datapath = [(shape_names[i], os.path.join(self.root, shape_names[i], shape_ids[split][i]) + '.txt') for i
                         in range(len(shape_ids[split]))]
        print('The size of %s data is %d' % (split, len(self.datapath)))

        self._cache = {}
        self.normal_small_knn = normal_small_knn
        self.normal_large_knn = normal_large_knn
        self.show_info = show_info

    def __len__(self):
        return len(self.datapath)

    def _get_item(self, index):
        name, ptfile = self.datapath[index]

        if self.show_info:
            klog.info(f'{name} {ptfile}')

        cls = self.classes[name]
        label = np.array([cls]).astype(np.int32)
        point_set = np.loadtxt(ptfile, delimiter=',').astype(np.float32)

        normal_samll_file = ptfile.replace('.txt', "-%02d.normals" % self.normal_small_knn)
        if self.show_info:
            klog.info(f'normal_samll_file: {normal_samll_file}')
        normal_small_set = np.loadtxt(normal_samll_file, delimiter=',').astype(np.float32)

        normal_large_file = ptfile.replace('.txt', "-%02d.normals" % self.normal_large_knn)
        if self.show_info:
            klog.info(f'normal_large_file: {normal_large_file}')
        normal_large_set = np.loadtxt(normal_large_file, delimiter=',').astype(np.float32)

        normal_set = (normal_small_set - normal_large_set) / 2
        point_set = np.concatenate([point_set[:, 0:3], normal_set], axis=1)
        # point_set = kcv3.farthest_point_sample(point_set, self.npoints)
        if self.split == 'train':
            point_set = kcv3.random_sample(point_set, self.npoints)
        else:
            point_set = point_set[:self.npoints, :]

        point_set[:, 0:3] = kcv3.pc_normalize(point_set[:, 0:3])

        if self.transform:
            point_set = self.transform(point_set)
        return point_set, label[0]

    def __getitem__(self, index):
        '''
        :param index:
        :return:  point_set, label
        '''
        return self._get_item(index)


if __name__ == '__main__':
    import laok.util.conf as kconf
    data = ModelnetNormalsDiff(kconf.get_conf('datasets').ModelnetNormals,  num_point=1024, show_info=True, cached=True)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/7/11 14:36:36

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import cv2
import numpy as np
from .cvt import swap_rb, keep_bgr, keep_gray
from .info import is_color, image_size
#===============================================================================
'''     
'''
#===============================================================================
__all__ = ['fix_cv_read',
           'read_image', 'read_color', 'read_gray',
           'write_image',
           'read_video', 'read_camera', 'VideoWriter',
           ]

####################    读取图像
def fix_cv_read():
    '''主要功能是修复 读取中文路径时候的问题,本质上就是替换 cv2.imread函数
    '''
    _org_read = cv2.imread
    cv2.imread = read_image

def read_image(filename, flags = cv2.IMREAD_COLOR):
    return cv2.imdecode(np.fromfile(filename, dtype=np.uint8), flags)

def read_color(filename, bgr=True):
    img = read_image(filename, cv2.IMREAD_COLOR)
    if not bgr:
        return swap_rb(img)
    return img

def read_gray(filename):
    return read_image(filename, cv2.IMREAD_GRAYSCALE)

def write_image(filename, data, params=None):
    name, ext = os.path.splitext(filename)
    retval, buf = cv2.imencode(ext, data, params)
    return buf.tofile(filename)

def write_jpg(filename, data, quality=95):
    return write_image(filename, data, [cv2.IMWRITE_JPEG_QUALITY, quality])

def write_png(filename, data, compression=1):
    return write_image(filename, data, [cv2.IMWRITE_PNG_COMPRESSION, compression])
####################    读取视频和相机
def _read_capture(fileOrId):
    try:
        cap = cv2.VideoCapture(fileOrId)
        while True:
            ret, frame = cap.read()
            if not ret:  # 如果正确读取帧，ret为True
                break
            yield frame
    finally:
        cap.release()

def read_video(video_file):
    yield from _read_capture(video_file)

def read_camera(camId = 0):
    yield from _read_capture(camId)


class VideoWriter:
    def __init__(self, filename, fourcc="mp4v", fps=24, frameSize=None, isColor=None):
        self.filename_ = filename
        self.frameSize_ = frameSize
        self.isColor_ = isColor
        self.fps_ = fps
        self.fourcc_ =  cv2.VideoWriter_fourcc(*fourcc)
        self.writer_ = None

    def _make_video(self, img):
        if self.frameSize_ is None:
            self.frameSize_ = image_size(img)

        if self.isColor_ is None:
            self.isColor_ = is_color(img)
        self.writer_ = cv2.VideoWriter(filename=self.filename_, fourcc=self.fourcc_,
                                      fps=self.fps_, frameSize=self.frameSize_, isColor=self.isColor_)

    def write(self, img):
        if self.writer_ is None:
            self._make_video(img)

        if self.isColor_ :
            img = keep_bgr(img)
        else:
            img = keep_gray(img)

        if image_size(img) != self.frameSize_:
            img = cv2.resize(img, dsize=(self.frameSize_[0], self.frameSize_[1]),
                             interpolation=cv2.INTER_LINEAR)
        self.writer_.write(img)

    def close(self):
        if self.writer_ is not None:
            self.writer_.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/5/22 17:32:21

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import logging
import logging.handlers
import functools
#===============================================================================
'''     
'''
#===============================================================================
__all__ = [ 'get_logger',
            'log_set_format',
            'log_add_rotate_file',
            'log_add_file',
            'log_add_console',
]

_def_name = 'laok'
_def_fmt = '[%(asctime)s][%(levelname)s][%(name)s]%(message)s'
_def_fmt_func_line = '[%(asctime)s][%(levelname)s][%(name)s][%(filename)s][%(funcName)s][%(lineno)d] %(message)s'

_logger_cache = {}
logging.basicConfig(format=_def_fmt)

def log_set_format(fmt=None, logger=None):
    if not logger:
        logger = get_logger()
    for hdl in logger.handlers:
        hdl.setFormatter(logging.Formatter(fmt if fmt else _def_fmt))


def log_add_rotate_file(filename=None, fmt = None, logger=None):
    if not logger:
        logger = get_logger()
    handler = logging.handlers.RotatingFileHandler(
        filename if filename else logger.name + ".log",
        maxBytes=1024 * 1024 * 1024 * 1,
        backupCount=5,
    )
    handler.setFormatter(logging.Formatter(fmt if fmt else _def_fmt))
    logger.addHandler(handler)
    return handler

def log_add_file(filename=None, fmt = None, logger=None):
    if not logger:
        logger = get_logger()
    handler = logging.FileHandler(filename if filename else logger.name + ".log")
    handler.setFormatter(logging.Formatter(fmt if fmt else _def_fmt))
    logger.addHandler(handler)
    return handler

def log_add_console(fmt = None, logger=None):
    if not logger:
        logger = get_logger()
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(fmt if fmt else _def_fmt))
    logger.addHandler(handler)
    return handler

def get_logger(name=None, level=logging.DEBUG):
    if name is None:
        name = _def_name

    if name in _logger_cache:
        return _logger_cache[name]

    log = logging.getLogger(name)
    _logger_cache[name] = log
    log.setLevel(level)
    return log






#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2025/5/15 14:11:25

@author: Liu Kuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
import torch
from tqdm import tqdm
# ===============================================================================
r'''
'''
# ===============================================================================


def test(model, loader, num_class=40, use_cpu=True):
    mean_correct = []
    class_acc = np.zeros((num_class, 3))
    classifier = model.eval()

    for j, (points, target) in tqdm(enumerate(loader), total=len(loader)):

        if not use_cpu:
            points, target = points.cuda(), target.cuda()

        points = points.transpose(2, 1)
        pred, _ = classifier(points)
        pred_choice = pred.data.max(1)[1]

        for cat in np.unique(target.cpu()):
            classacc = pred_choice[target == cat].eq(target[target == cat].long().data).cpu().sum()
            class_acc[cat, 0] += classacc.item() / float(points[target == cat].size()[0])
            class_acc[cat, 1] += 1

        correct = pred_choice.eq(target.long().data).cpu().sum()
        mean_correct.append(correct.item() / float(points.size()[0]))

    class_acc[:, 2] = class_acc[:, 0] / class_acc[:, 1]
    class_acc = np.mean(class_acc[:, 2])
    instance_acc = np.mean(mean_correct)

    return instance_acc, class_acc


val = '-0.24453,-0.570776,-0.78385'
x,y,z = [float(x) for x in val.split(',')]
print('len=', x*x + y*y + z*z)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/27 14:18:09

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import os.path as opath
from pathlib import Path
import shutil
#===============================================================================
#
#===============================================================================
__all__ = [ 'path_rename', 'path_native', 'path_unix', 'path_win', 'path_relative',
            'path_parent', 'path_make', 'path_make_parent',
            'path_join', 'path_abs', 'path_exist',
            'file_size', 'path_is_file', 'path_is_dir',
            'path_basename', 'path_stem', 'path_ext','path_add_ext', 'path_rm_ext',
            'path_replace_ext', 'path_replace_stem', 'path_replace_basename', 'path_replace_parent',
            'dirs_delete', 'file_delete', 'path_move',
            'path_deco_stem',
            ]

def path_rename(old_name, new_name):
    os.rename(old_name, new_name)

def path_native(fpath):
    return opath.normpath(fpath).replace('\\', '/').replace('/', os.sep)

def path_unix(fpath):
    return opath.normpath(fpath).replace('\\', '/')

def path_win(fpath):
    return opath.normpath(fpath).replace('/', '\\')

def path_relative(fpath, parent=None):
    return path_unix(os.path.relpath(fpath, parent))

def path_parent(fpath, level=1):
    '''fpath: 输入路径名字
       ret: 返回父路径
    '''
    p_path = fpath
    for i in range(0, level):
        p_path = path_unix(opath.split(p_path)[0])
    return p_path

def path_make(fpath):
    os.makedirs(fpath, exist_ok=True)
    return fpath

def path_make_parent(fpath):
    parent = path_parent(fpath)
    return path_make(parent)

def path_join(*args, make_parent=False):
    fpath = os.path.join(*args)
    if make_parent:
        path_make_parent(fpath)
    return path_unix(fpath)

def path_abs(fpath):
    return path_unix(os.path.abspath(fpath))

def path_exist(fpath):
    return fpath and os.path.exists(fpath)

def file_size(fpath):
    return os.path.getsize(fpath)

def path_is_file(fpath):
    return fpath and os.path.isfile(fpath)

def path_is_dir(fpath):
    return fpath and os.path.isdir(fpath)

def path_basename(fpath):
    return os.path.basename(fpath)

def path_stem(fpath):
    return Path(fpath).stem

def path_ext(fpath, need_dot = True):
    '''file_path: 输入路径名字
       ret: 返回后缀
    '''
    fname = opath.split(fpath)[1]
    ext = opath.splitext(fname)[1]
    if ext and not need_dot:
        return ext[1:]
    return ext

def path_add_ext(fpath, ext):
    if ext.startswith('.'):
        return fpath + ext
    else:
        return fpath + "." + ext

def path_rm_ext(fpath):
    return path_unix(opath.splitext(fpath)[0])

def path_replace_ext(file_path, ext):
    if isinstance(ext, str) and not ext.startswith('.'):
        ext = '.' + ext
    new_file = opath.splitext(file_path)[0] + ext
    return path_unix(new_file)

def path_replace_stem(file_path, basename):
    _dir, _fname = opath.split(file_path)
    _basename, _ext = opath.splitext(_fname)
    return path_join(_dir, basename + _ext)

def path_deco_stem(file_path, preffix='', suffix=''):
    _dir, _fname = opath.split(file_path)
    _basename, _ext = opath.splitext(_fname)
    return path_join(_dir, preffix + _basename + suffix + _ext)

def path_replace_basename(file_path, filename):
    _dir, _fname = opath.split(file_path)
    return path_join(_dir, filename)

def path_replace_parent(file_path, parent):
    _dir, _fname = opath.split(file_path)
    return path_join(parent, _fname)

def dirs_delete(dirpath):
    shutil.rmtree(dirpath)

def file_delete(fpath):
    os.remove(fpath)

def path_move(old, new):
    shutil.move(old, new)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/16 10:07:05

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os, pathlib
from .fpath import path_join
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['dirs_cur', 'dirs_under',
           'files_cur', 'files_under',
           'fname_endswith', 'path_sep'
           ]

def path_sep(val, sep=';'):
    if val is None:
        return ()
    if isinstance(val, (list,tuple)):
        return val
    if isinstance(val, pathlib.Path):
        return str(val)
    if isinstance(val, str):
        return val.split(sep)
    return val

def fname_endswith(fname, suffixes):
    if suffixes is None:
        return True
    _low_suffix_list = [s.lower() for s in path_sep(suffixes)]
    _low_fname = fname.lower()
    return any( (_low_fname.endswith(s) for s in _low_suffix_list) )

####################    通用遍历路径

def dirs_cur(dir_names, suffixes = None):
    '''
    dir_names: 路径,支持多个
    suffixes: 后缀列表,支持多个;用于筛选路径
    ret: 返回生成器
    '''
    for dir_name in path_sep(dir_names):
        for fname in os.listdir(dir_name):
            fpath = path_join(dir_name, fname)
            if os.path.isdir(fpath):
                if fname_endswith(fpath, suffixes):
                    yield fpath

def dirs_under(dir_names, suffixes = None):
    '''
    dir_names: 路径,支持多个
    suffixes: 后缀列表,支持多个;用于筛选路径
    ret: 返回生成器
    '''
    for dir_name in path_sep(dir_names):
        for fdir_name, sub_fdirs, files in os.walk(dir_name):
            for fname in sub_fdirs:
                if fname_endswith(fname, suffixes):
                    yield path_join(fdir_name, fname)

def files_under(dir_names, suffixes = None):
    '''
    dir_names: 路径,支持多个
    suffixes: 后缀列表,支持多个;用于筛选路径
    ret: 返回生成器
    '''
    for dir_name in path_sep(dir_names):
        for fdir_name, _sub_fdirs, files in os.walk(dir_name):
            for fname in files:
                if fname_endswith(fname, suffixes):
                    yield path_join(fdir_name, fname)

def files_cur(dir_names, suffixes = None):
    '''
    dir_names: 路径,支持多个
    suffixes: 后缀列表,支持多个;用于筛选路径
    ret: 返回生成器
    '''
    for dir_name in path_sep(dir_names):
        for fname in os.listdir(dir_name):
            fpath = path_join(dir_name, fname)
            if os.path.isfile(fpath):
                if fname_endswith(fname, suffixes):
                    yield fpath



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/7/12 14:48:16

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2

# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['contour_extract_external', 'contour_extract_list', 'contour_extract_twolevel',

           ]

def contour_extract_external(img, method = cv2.CHAIN_APPROX_SIMPLE):
    contours, hierarchy = cv2.findContours(img, mode=cv2.RETR_EXTERNAL, method=method)
    return contours

def contour_extract_list(img, method = cv2.CHAIN_APPROX_SIMPLE):
    contours, hierarchy = cv2.findContours(img, mode=cv2.RETR_LIST, method=method)
    return contours

def contour_extract_twolevel(img, method = cv2.CHAIN_APPROX_SIMPLE):
    contours, hierarchy = cv2.findContours(img, mode=cv2.RETR_CCOMP, method=method)
    return contours

def contour_statis_area(contours):
    pass

def contour_statis_moment(contours):
    pass

def contour_fit_poly(contours):
    pass

def contour_fit_convex_hull(contours):
    pass

def contour_fit_ratate_rect(contours):
    pass

def contour_fit_bouding_rect(contours):
    pass

def contour_fit_min_enclose_circle(contours):
    pass

def contour_fit_ellipse(contours):
    pass

def contour_fit_line(contours):
    pass

def contour_filter_point_count():
    pass

def contour_filter_area():
    pass

def contour_filter_bounding_rect():
    pass














#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/3 11:00:32

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import laok
import laok.ext.requests_ as kreq
import laok.ext.Crypto_ as kcrypto
from collections import defaultdict
from tqdm import tqdm
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = [
            'ts_data_list_gen',
            'ts_data_list_filter',
            'ts_download',
            'ts_download_decrpyt',
            'ts_data_list_download',
            'ts_list_merge',
            'ts_download_m3u8'
           ]


# 下载 ts data list
# 存储为字典列表,内容有 url,key,iv,file
def ts_data_list_gen(url, del_cache=False):
    ts_data_list = []
    method = 'NONE'
    iv = b'0000000000000000'
    key = b''

    #保存 index文件
    index_file = kreq.req_save_auto(url, skip_exist=True)

    #保存 url
    url_file = laok.path_replace_basename(index_file, 'url.txt')
    laok.file_write_text(url_file, url, skip_exist=True)

    #读取index文件
    for line in laok.file_read_lines(index_file, skip_empty=True):
        if line.startswith('#EXT-X-KEY'):
            kdata = [v.strip() for v in line[11:].split(',')]
            method = 'NONE'
            iv = b'0000000000000000'
            key = b''
            for d in kdata:
                if d.startswith('METHOD='):
                    method = d[7:]
                elif d.startswith('URI='):
                    uri = laok.url_join(url, d[4:].strip('""') )
                    key_file = kreq.req_save_auto(uri, skip_exist=True)
                    key = laok.file_read_bin(key_file)
                    if del_cache:
                        laok.file_delete(key_file)
        elif line.startswith('#EXT-X-ENDLIST'):
            break
        elif line[0] != '#':
            uri = laok.url_join(url, line)
            ts_data_list.append({ 'url': uri,
                                  'key': key,
                                  'iv': iv,
                                  'method': method,
            })

    if del_cache:
        laok.file_delete(index_file)
        laok.file_delete(url_file)

    return ts_data_list

# 根据路径前缀过滤 插入的广告
def ts_data_list_filter(ts_list):
    prefix_dict = defaultdict(list)
    for ts in ts_list:
        prefix_dict[laok.path_parent(ts['url'])].append(ts)
    return max(prefix_dict.values(), key= lambda x : len(x) )


# 下载解密的 ts文件
def ts_download_decrpyt(url, key, iv):
    ts_file = laok.url_make_file(url)
    if laok.path_exist(ts_file):
        return ts_file
    kreq.req_save(ts_file, url)
    if key:
        kcrypto.aes_decrypt_cbc_file_(ts_file, key, iv)
    return ts_file

# 下载未解密的 ts文件
def ts_download(url):
    ts_file = laok.url_make_file(url)
    return kreq.req_save(ts_file, url, skip_exist=True)

# 下载 ts data list
# 从ts_data_list_gen/ts_data_list_filter 获取结果
def ts_data_list_download(ts_data_list, queue_size=100):
    pbar = tqdm(total=len(ts_data_list)+1)
    def download_ts_data(ts_data):
        pbar.set_description(ts_data['url'])
        pbar.update(1)
        return ts_download_decrpyt(ts_data['url'], ts_data['key'], ts_data['iv'])
    return laok.paral_seq(ts_data_list, download_ts_data, queue_size)

# 合并 ts_list 文件,并保存到 dst_file
# del_cache 是否删除缓冲文件
def ts_list_merge(ts_list, dst_file, del_cache=False):
    if len(ts_list) < 1:
        return

    dst_file = laok.path_replace_ext(dst_file, '.mp4')
    laok.file_list_merge(ts_list, dst_file)
    if not del_cache:
        return

    #删除所有ts
    for ts in ts_list:
        laok.file_delete(ts)

#下载 m3u8文件
def ts_download_m3u8(url, dst_file, queue_size=100, del_cache=False):
    ts_data_list = ts_data_list_gen(url, del_cache=del_cache)
    ts_data_list = ts_data_list_filter(ts_data_list)
    ts_list = ts_data_list_download(ts_data_list, queue_size=queue_size)
    ts_list_merge(ts_list, dst_file, del_cache=del_cache)

def convert_m3u8_mp4(index_file):

    pass

if __name__ == '__main__':
    import fire as _fire
    _fire.Fire(ts_download_m3u8)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/7/12 15:29:37

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ =['bin_otsu', 'bin_triangle', 'bin_mean',
          'bin_threshold',
          'bin_adapt_mean', 'bin_adapt_gauss']

def bin_otsu(img, scale=1.0, inv=False):
    flag = cv2.THRESH_BINARY_INV if inv else cv2.THRESH_BINARY
    th, ret = cv2.threshold(img, 0, 255, cv2.THRESH_OTSU + flag)
    if scale == 1.0:
        return ret
    th, ret = cv2.threshold(img, th*scale, 255, flag)
    return ret

def bin_triangle(img, scale=1.0, inv=False):
    flag = cv2.THRESH_BINARY_INV if inv else cv2.THRESH_BINARY
    th, ret = cv2.threshold(img, 0, 255, cv2.THRESH_TRIANGLE + flag)
    if scale == 1.0:
        return ret
    th, ret = cv2.threshold(img, th*scale, 255, flag)
    return ret

def bin_mean(img, scale=1.0, inv=False):
    flag = cv2.THRESH_BINARY_INV if inv else cv2.THRESH_BINARY
    value = cv2.mean(img)
    th, ret = cv2.threshold(img, value[0]*scale, 255, flag)
    return ret

def bin_threshold(img, value=128, inv=False):
    flag = cv2.THRESH_BINARY_INV if inv else cv2.THRESH_BINARY
    th, ret = cv2.threshold(img, value, 255, flag)
    return ret

def bin_adapt_mean(img, blockSize=11, C=2, inv=False):
    flag = cv2.THRESH_BINARY_INV if inv else cv2.THRESH_BINARY
    return cv2.adaptiveThreshold(img, maxValue=255,
                                 adaptiveMethod=cv2.ADAPTIVE_THRESH_MEAN_C,
                                 thresholdType=flag, blockSize=blockSize, C=C)

def bin_adapt_gauss(img, blockSize=11, C=2, inv=False):
    flag = cv2.THRESH_BINARY_INV if inv else cv2.THRESH_BINARY
    return cv2.adaptiveThreshold(img, maxValue=255,
                                 adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                 thresholdType=flag, blockSize=blockSize, C=C)


